"""
xOpt = [ 2.35342485  3.32421007]
fOpt = 7.82327859
"""
from DesOptPy import OptimizationProblem
import numpy as np
import time


class Cantilever:
    e = 2.9e7
    r = 40000.0
    fx = 500.0
    fy = 1000.0
    w = 1.8
    t = 1.0
    L = 100.0
    D0 = 2.2535
    area = None
    stress = None

    def calc(self):
        self.area = self.w * self.t
        self.stress = (
            600 * self.fy / self.w / self.t ** 2
            + 600 * self.fx / self.w ** 2 / self.t
        )
        D1 = 4.0 * pow(self.L, 3) / self.e / self.area
        D2 = pow(self.fy / self.t ** 2, 2) + pow(self.fx / self.w ** 2, 2)
        D3 = D1 / np.sqrt(D2) / self.D0
        self.D4 = D1 * np.sqrt(D2) / self.D0


Prob1 = OptimizationProblem(Cantilever)
Prob1.RemoveRunFolder = False
Prob1.Monitoring = True
Prob1.Primal = 'calc'
Prob1.x = ['w', 't']
Prob1.x0 = [1.8, 1.0]
Prob1.xL = [1.0, 1.0]
Prob1.xU = [4.0, 4.0]
Prob1.f = ['area']
Prob1.g = ['stress', 'D4']
Prob1.gLimit = [40000, 1]
Prob1.gNorm = [True, False]
Prob1.Alg = 'NLPQLP'
Prob1.optimize()
Prob1.plotConvergence()
Prob1.plotBeforeAfter()


Prob2 = OptimizationProblem(Cantilever)
Prob2.Primal = 'calc'
Prob2.x = ['w']
Prob2.x0 = [1.8]
Prob2.xL = [1.0]
Prob2.xU = [100.0]
Prob2.f = ['area']
Prob2.g = ['stress', 'D4']
Prob2.gType = ['upper', 'upper']
Prob2.gNorm = [True, False]
Prob2.gLimit = [40000, 1]
Prob2.Alg = 'NLPQLP'
Prob2.optimize()
Prob2.plotConvergence()
Prob2.plotBeforeAfter()

Prob3 = OptimizationProblem(Cantilever)
Prob3.Primal = 'calc'
Prob3.x = ['t']
Prob3.x0 = [1.8]
Prob3.xL = [1.0]
Prob3.xU = [100.0]
Prob3.xNorm = [True]
Prob3.xType = ['continuous']
Prob3.f = ['area']
Prob3.g = ['stress', 'D4']
Prob3.gType = ['upper', 'upper']
Prob3.gNorm = [True, False]
Prob3.gLimit = [40000, 1]
Prob3.Alg = 'NLPQLP'
Prob3.optimize()
Prob3.plotConvergence()
Prob3.plotBeforeAfter()



from DesOptPy import OptimizationProblem
from scipy.optimize import rosen, rosen_der
import numpy as np


class Rosenbrock:
    def calc(self):
        self.obj = rosen(self.x)

    def calcSens(self):
        self.objNabla = rosen_der(self.x)


ProbRosenbrockSens = OptimizationProblem(Rosenbrock)
ProbRosenbrockSens.RunFolder = True
nx = 20
ProbRosenbrockSens.x = 'x'
ProbRosenbrockSens.x0 = (
    np.ones(
        nx,
    )
    * +2
)
ProbRosenbrockSens.xL = (
    np.ones(
        nx,
    )
    * -5
)
ProbRosenbrockSens.xU = (
    np.ones(
        nx,
    )
    * +5
)
ProbRosenbrockSens.xNorm = [True] * nx
ProbRosenbrockSens.xType = ['continuous'] * nx
ProbRosenbrockSens.f = ['obj']
ProbRosenbrockSens.fNorm = [False]
ProbRosenbrockSens.Alg = 'NLPQLP'
ProbRosenbrockSens.Primal = 'calc'
ProbRosenbrockSens.Sensitivity = 'calcSens'
ProbRosenbrockSens.fNabla = ['objNabla']
ProbRosenbrockSens.TablesPlots = True
ProbRosenbrockSens.optimize()
ProbRosenbrockSens.plotConvergence()
ProbRosenbrockSens.plotBeforeAfter()



"""
Function aggregation with
    Kreisselmeier-Steinhauser (KS)
    p-Norm (pNorm)
    Kreisselmeier-Steinhauser with offset to avoid data overflow (KS2)
    p-Norm with offset to avoid data overflow (pNorm2)
    Induced exponential (IE)
    Induced power (IP)

Recommendation for use: KS2 with rho>100


Literature:
    G. Kreisselmeier and R. Steinhauser (1979) Systematic control design by
    optimizing a vector performance index. In International Federation of
    Active Controls Syposium on Computer-Aided Design of Control Systems,
    Zurich, Switzerland.

Second version of the functions as described in
    A. B. Lambe, J. R. R. A. Martins, and G. J. Kennedy. An evaluation of
    constraint aggregation strategies for wing boxmass minimization. Structural
    and Multidisciplinary Optimization, 55(1):257–277, January 2017
Was the KS version of this first published in???
    Martins and Poon (2005) On structural optimization using constraint
    aggregation 6th World Congress on Structural and Multidisciplinary
    Optimization, Rio de Janeiro, Brazil.

Induced exponential aggregation introduced by
    Kennedy (2015) Strategies for adaptive optimization with aggregation
    constraints using interior-point methods, Computers & Structures, 153:
    217-229.
and
    Kennedy and Hicken (2015) Improved constraint-aggregation methods. Computer
    Methods in Applied Mechanics and Engineering 289:332-354.

Papers can be found here:
    http://gkennedy.gatech.edu/publications/
    http://mdolab.engin.umich.edu/biblio


Note:
p-norm and induced power aggregation do not work with negative numbers or mix
of positive and negative numbers(minimum problem, i.e. frequencies)


TODO implement sensitivities for IP and IE
"""
import numpy as np


def KS(f, rho=200):
    """
    Return returns scalar KS function from vector f.

    # rho represents the closeness to the limit [1, 100] other papers [5, 50]
    """
    rho = float(rho)
    return 1 / rho * (np.log(np.sum(np.exp(rho * f))))


def KSSens(f, dfdx, rho=200):
    """
    Return vector sensitvities of KS function from vector f and matrix dfdx.
    rho represents the closeness to the limit [1, 100] other papers [5, 50]
    """
    rho = float(rho)
    return 1 / np.sum(np.exp(rho * f)) * np.exp(rho * f) @ dfdx


def KS2(f, rho=200):
    """
    Return returns scalar KS function from vector f.
    rho represents the closeness to the limit [1, 100] other papers [5, 50]
    """
    rho = float(rho)
    return np.max(f) + 1 / rho * np.log(np.sum(np.exp(rho * (f - np.max(f)))))


def KS2Sens(f, dfdx, rho=200):
    """
    Return vector sensitvities of KS function from vector f and matrix dfdx.
    rho represents the closeness to the limit [1, 100] other papers [5, 50]
    """
    rho = float(rho)
    return (
        1
        / np.sum(np.exp(rho * (f - np.max(f))))
        * np.exp(rho * (f - np.max(f)))
        @ dfdx
    )


def pNorm(f, p=200):
    """Return returns scalar p-norm from vector f."""
    p = float(p)
    return np.sum(f ** p) ** (1 / p)


def pNormSens(f, dfdx, p=200):
    """Return vector sensitvities of p-norm from vector f and matrix dfdx."""
    p = float(p)
    return np.sum(f ** p) ** (1 / p - 1) * (f ** (p - 1)) @ dfdx


def pNorm2(f, p=200):
    """Return scalar p-norm from vector f."""
    p = float(p)
    return np.max(f) * np.sum((f / np.max(f)) ** p) ** (1 / p)


def pNorm2Sens(f, dfdx, p=200):
    """Return vector sensitvities of p-norm from vector f and matrix dfdx."""
    p = float(p)
    return (
        np.sum((f / np.max(f)) ** p) ** (1 / p - 1)
        * ((f / np.max(f)) ** (p - 1))
        @ dfdx
    )


def IE(f, rho=200):
    """Return scalar induced exponential aggregation from vector f."""
    rho = float(rho)
    return np.sum(f * np.exp(rho * f)) / np.sum(np.exp(rho * f))


def IP(f, rho=200):
    rho = float(rho)
    return np.sum(f ** (rho + 1)) / np.sum(f ** rho)


if __name__ == '__main__':
    print('')
    print('Testing of functions:')
    print('')
    p = rho = 200
    print('-' * 50)
    print('Example 1 - simple example, small values, small number of values')
    print('-' * 50)
    f = np.array([1, 2, 3])
    dfdx = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]]).T
    dfdx[-1, -1] = 0
    print('fKS:', KS(f, rho))
    print('nabla_fKS:', KSSens(f, dfdx, rho))
    print('fp:', pNorm(f, p))
    print('nabla_fp:', pNormSens(f, dfdx, p))
    print('fKS2:', KS2(f, rho))
    print('nabla_fKS2:', KS2Sens(f, dfdx, rho))
    print('fp2:', pNorm2(f, p))
    print('nabla_fp2:', pNorm2Sens(f, dfdx, p))
    print('fIE:', IE(f, rho))
    print('fIP:', IP(f, rho))
    print('')
    print('True values:')
    print('fMax :', np.max(f))
    print('fMaxNabla :', dfdx[:, -1])
    print('-' * 50)
    print('')
    print('-' * 50)
    print('Example 2 - large values, large number of values')
    print('-' * 50)
    nr2 = 100000
    nx2 = 3
    fVal = 10000
    dfdxVal = 10000
    f = np.linspace(1, fVal, nr2)
    dfdx = np.linspace(1, dfdxVal, nr2 * nx2).reshape(nr2, nx2)
    dfdx[0:-1, -1] = 0.0
    print('fKS:', KS(f, rho))
    print('nabla_fKS:', KSSens(f, dfdx, rho))
    print('fp:', pNorm(f, p))
    print('nabla_fp:', pNormSens(f, dfdx, p))
    print('fKS2:', KS2(f, rho))
    print('nabla_fKS2:', KS2Sens(f, dfdx, rho))
    print('fp2:', pNorm2(f, p))
    print('nabla_fp2:', pNorm2Sens(f, dfdx, p))
    print('fIE:', IE(f, rho))
    print('fIP:', IP(f, rho))
    print('')
    print('True values:')
    print('fMax :', np.max(f))
    print('fMaxNabla :', dfdx[:, -1])
    print('-' * 50)
    print('')
    print('-' * 50)
    print('Example 3 - negative values')
    print('-' * 50)
    f = np.array([1, 2, 3])
    dfdx = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]]).T
    dfdx[-1, -1] = 0
    print('fKS:', -KS(-f, rho))
    print('nabla_fKS:', -KSSens(-f, -dfdx, rho))
    print('fp:', -pNorm(-f, p))
    print('nabla_fp:', -pNormSens(-f, -dfdx, p))
    print('fKS2:', -KS2(-f, rho))
    print('nabla_fKS2:', -KS2Sens(-f, -dfdx, rho))
    print('fp2:', -pNorm2(-f, p))
    print('nabla_fp2:', -pNorm2Sens(-f, -dfdx, p))
    print('fIE:', -IE(-f, rho))
    print('fIP:', -IP(-f, rho))
    print('')
    print('True values:')
    print('fMin:', np.min(f))
    print('fMinNabla :', dfdx[:, 0])
    print('-' * 50)
    print('')
    print('-' * 50)
    print('Example 4 - pseudo topology optimization, very large sparse matrix')
    from scipy import sparse as sp

    print('-' * 50)
    nr = 1000000
    nx = 1000000
    fVal = 1000
    dfdxVal = 10000
    fTop = np.linspace(1, fVal, nr)
    dfdxTop = sp.diags(fTop)
    # dfdxTop = np.diag(fTop)
    print('fKS2:', KS2(fTop, rho))
    dfdxTopKS2 = KS2Sens(fTop, dfdxTop, rho)
    # print("max nabla_fKS2:", np.max(dfdxTopKS2))
    print('where nabla_fKS2>1:', np.argwhere(dfdxTopKS2 > 1e0))
    print('nabla_fKS2:', dfdxTopKS2[np.argwhere(dfdxTopKS2 > 1e0)])
    print('max nabla_fKS2:', np.max(dfdxTopKS2))
    print('fp2:', pNorm2(fTop, p))
    dfdxToppNorm2 = KS2Sens(fTop, dfdxTop, rho)
    print('n(nabla_fp2)>1.0:', np.size(np.argwhere(dfdxToppNorm2 > 1e0)))
    print('max nabla_fpNorm2:', np.max(dfdxToppNorm2))
    print('')
    print('True values:')
    print('fMax :', np.max(fTop))
    print('fNabla > 1e0: [' + str(nx - 1) + ']')
    print('fNabla:', dfdxTop.tocsr()[nr - 1, nx - 1])
    print('-' * 50)
    print('')
    print('-' * 50)
    print('Example 5 - large values, large numbers, range of exponent')
    print('-' * 50)
    from matplotlib import pylab as plt

    nr2 = 100000
    nx2 = 3
    fVal = 10000
    dfdxVal = 10000
    f = np.linspace(1, fVal, nr2)
    dfdx = np.linspace(1, dfdxVal, nr2 * nx2).reshape(nx2, nr2)
    dfdx[0:-1, -1] = 0.0
    rhoList = np.logspace(-3, 5, 1001)

    fKS = []
    fpNorm = []
    fKS2 = []
    fpNorm2 = []
    fIE = []
    fIP = []
    for rhoi in rhoList:
        fKS.append(KS(f, rhoi))
        fKS2.append(KS2(f, rhoi))
        fpNorm.append(pNorm(f, rhoi))
        fpNorm2.append(pNorm2(f, rhoi))
        fIE.append(IE(f, rhoi))
        fIP.append(IP(f, rhoi))
    plt.semilogx(
        [10 ** -3, 10 ** 5], [fVal, fVal], 'k', linewidth=0.5, label='true'
    )
    plt.semilogx(rhoList, fKS, 'r', linestyle='dotted', label='KS')
    plt.semilogx(rhoList, fKS2, 'r', linestyle=(0, (5, 10)), label='KS 2')
    plt.semilogx(rhoList, fpNorm, 'b', linestyle='dotted', label='p-norm')
    plt.semilogx(
        rhoList, fpNorm2, 'b', linestyle=(0, (5, 10)), label='p-norm 2'
    )
    plt.semilogx(rhoList, fIE, 'g-', label='IE')
    plt.semilogx(rhoList, fIP, 'y-', label='IP')
    plt.xlim((10 ** -3, 10 ** 5))
    plt.ylim((8000, 12000))
    plt.xlabel('exponent value $\\rho$ (KS) or $p$ (p-norm)')
    plt.ylabel('approximated maximum value $\\max(f)$')
    plt.legend(frameon=False, loc='center left', bbox_to_anchor=(1, 0.5))
    plt.title('Example 5: large values, large number of values')
    plt.show()

    print(
        """
Explanation of above:
    Original functions do not work with large numbers of values or large values
    p-Norm has problems with sensitvitivies with large difference in values
    p-Norm does not work with negative numbers
    Generally KS2 works well in all cases
          """
    )



"""
f*=0
x*=1
"""
from DesOptPy import OptimizationProblem
import numpy as np
from scipy.optimize import rosen, rosen_der


class Rosenbrock:
    def calc(self):
        self.obj = rosen(self.x)

    def calcSens(self):
        self.objNabla = rosen_der(self.x)


ProbRosenbrockSens = OptimizationProblem(Rosenbrock)
ProbRosenbrockSens.Primal = 'calc'
ProbRosenbrockSens.Sensitivity = 'calcSens'
nx = 10
ProbRosenbrockSens.x = 'x'
ProbRosenbrockSens.x0 = (
    np.ones(
        nx,
    )
    * 3
)
ProbRosenbrockSens.xL = (
    np.ones(
        nx,
    )
    * -5
)
ProbRosenbrockSens.xU = (
    np.ones(
        nx,
    )
    * +5
)
ProbRosenbrockSens.f = ['obj']
ProbRosenbrockSens.fNabla = ['objNabla']
ProbRosenbrockSens.Alg = 'SLSQP'
ProbRosenbrockSens.optimize()
ProbRosenbrockSens.plotConvergence()



"""
http://cvxopt.org/userguide/index.html
http://cvxopt.org/userguide/solvers.html
"""
import cvxopt


def CvxOptOpt(self, x0, xL, xU, SysEq):
    def F(x=None, z=None):

        return f, Df, H

    cvxopt.solvers.cp(F)



from DesOptPy.DesOptPy import *
from DesOptPy.tools import *
from DesOptPy.plotting import *

# from DesOptPy.FunctionAggregation import *


__title__ = 'DESign OPTimization in PYthon'
__shorttitle__ = 'DesOptPy'
__version__ = '2.0 development'
# __all__ = ["DesOpt"]
__author__ = 'E. J. Wehrle'
__copyright__ = 'Copyright 2021 E. J. Wehrle'
__email__ = 'Erich.Wehrle(a)unibz.it'
__url__ = 'github.com/edub/DesOptPy2'
# TODO decide on license for new version
__license__ = 'TBD'  # "GNU Lesser General Public License"


def PrintDesOptPy():
    print()
    print(__title__ + ' - ' + __shorttitle__)
    print('Version:    ' + __version__)
    print('License:    ' + __license__)
    print('Internet:   ' + __url__)
    print('\n')


PrintDesOptPy()



def normalize(x, xL, xU):
    return (x - xL) / (xU - xL)


def denormalize(xNorm, xL, xU):
    return xNorm * (xU - xL) + xL


if __name__ == '__main__':
    print('testing')
    import numpy as np

    x0 = np.array([9, 900])
    xL = np.array([10, 10])
    xU = np.array([1000, 1000])
    xNorm = normalize(x0, xL, xU)
    print(xNorm)
    x = denormalize(xNorm, xL, xU)
    print(x)



# -*- coding: utf-8 -*-
"""
Title:              RangeConstraint.py
Units:              -
Author:             E.J. Wehrle, V. Gufler
Date:               September 21, 2018
-------------------------------------------------------------------------------
Description:
A function for band constraints, as in the constraining of
resonance frequency bands.
-------------------------------------------------------------------------------
Future work:
TODO Add sensitivities of ballistic constraint function
TODO Add plotting capability of constraint functions

Warnings:
RangeConstraint.py:94: RuntimeWarning: overflow encountered in exp
  np.exp(c*(omega[jj]+d-omegaBands[ii][0])))

RangeConstraint.py:116: RuntimeWarning: overflow encountered in exp
  dgdx[ig, :] = c/a*domegadx[jj]*(1-np.exp(c*(omega[jj]+d -
-------------------------------------------------------------------------------
"""
import numpy as np
from scipy.special import lambertw


def BandConParabolicFn(omega, omegaBands, c=2e-5, norm=False):
    g = np.zeros((np.shape(omegaBands)[0] * np.size(omega)))
    ig = 0
    if np.size(omega) == 1:
        omega = [omega]
    for jj in range(np.size(omega)):
        for ii in range(np.shape(omegaBands)[0]):
            if norm:
                g[ig] = (
                    c
                    * (omega[jj] / omegaBands[ii][0] - 1)
                    * (1 - omega[jj] / omegaBands[ii][1])
                )
            else:
                g[ig] = (
                    c
                    * (omegaBands[ii][0] - omega[jj])
                    * (omega[jj] - omegaBands[ii][1])
                )
            ig += 1
    return g


def ConSens(omega, domegadx, omegaL, omegaU, c):
    return (
        -c * omegaU * (-1 + omega / omegaL) * domegadx / omega ** 2
        + c * (omegaU / omega - 1) * domegadx / omegaL
    )


def BandConParabolicSensFn(omega, domegadx, omegaBands, c=2e-5, norm=False):
    dgdx = np.zeros(np.size(omega) * np.shape(omegaBands)[0])
    ig = 0
    for jj in range(np.size(omega)):
        for ii in range(np.shape(omegaBands)[0]):
            if norm:
                dgdx[ig] = ConSens(
                    omega[jj],
                    domegadx[jj],
                    omegaBands[ii, 0],
                    omegaBands[ii, 0],
                    (omegaBands[ii, 1] + omegaBands[ii, 0]) / 2.0,
                )
            else:
                dgdx[ig] = (
                    c * omegaBands[ii][0] * domegadx[ii]
                    + c * omegaBands[ii][1] * domegadx[ii]
                    + c * 2 * omega[ii] * domegadx[ii]
                )
            ig += 1
    return dgdx


def BandConSawToothFn(omega, omegaBands, norm=True, c=[]):
    g = np.zeros((np.shape(omegaBands)[0] * np.size(omega)))
    ig = 0
    if np.size(omega) == 1:
        omega = [omega]
    for jj in range(np.size(omega)):
        for ii in range(np.shape(omegaBands)[0]):
            if norm:
                if omega[jj] < omegaBands[ii][1]:
                    g[ig] = (
                        (omega[jj] / omegaBands[ii][0] - 1)
                        * omegaBands[ii][0]
                        / (omegaBands[ii][1] - omegaBands[ii][0])
                    )
                else:
                    g[ig] = (
                        (1 - omega[jj] / omegaBands[ii][1])
                        * omegaBands[ii][0]
                        / (omegaBands[ii][1] - omegaBands[ii][0])
                    )
            else:
                if omega[jj] < omegaBands[ii][1]:
                    g[ig] = omega[jj] - omegaBands[ii][0]
                else:
                    g[ig] = omegaBands[ii][1] - omega[jj]
            ig += 1
    return g


def BandConBallisticFn(
    omega, omegaBands, a=1e-8, b=0, norm=True, infFilter=False
):
    g = np.zeros((np.shape(omegaBands)[0] * np.size(omega)))
    ig = 0
    if np.size(omega) == 1:
        omega = [omega]
    for jj in range(np.size(omega)):
        for ii in range(np.shape(omegaBands)[0]):
            c = (
                (-1 - 2 * b)
                / (omegaBands[ii][1] - omegaBands[ii][0])
                * (
                    lambertw(-1 / np.exp(1 + a), -1)
                    - lambertw(-1 / np.exp(1 + a), 0)
                ).real
            )
            d = 1 / c * (-lambertw(-1 / np.exp(1 + a), b) - (1 + a)).real
            if norm:
                g[ig] = (
                    1
                    / a
                    * (
                        (1 + a)
                        + c * (omega[jj] + d - omegaBands[ii][0])
                        - np.exp(c * (omega[jj] + d - omegaBands[ii][0]))
                    )
                )
            else:
                g[ig] = (
                    (1 + a)
                    + c * (omega[jj] + d - omegaBands[ii][0])
                    - np.exp(c * (omega[jj] + d - omegaBands[ii][0]))
                )
            ig += 1
            if infFilter:
                g[g < -infFilter] = -infFilter
    return g


def BandConBallisticSensFn(
    omega, domegadx, omegaBands, a=1e-8, b=0, norm=True, infFilter=False
):

    dgdx = np.zeros(
        [np.size(omega) * np.shape(omegaBands)[0], np.shape(domegadx)[1]]
    )

    ig = 0
    for jj in range(np.size(omega)):
        for ii in range(np.shape(omegaBands)[0]):
            c = (
                (-1 - 2 * b)
                / (omegaBands[ii][1] - omegaBands[ii][0])
                * (
                    lambertw(-1 / np.exp(1 + a), -1)
                    - lambertw(-1 / np.exp(1 + a), 0)
                ).real
            )
            d = 1 / c * (-lambertw(-1 / np.exp(1 + a), b) - (1 + a)).real
            if norm:
                dgdx[ig, :] = (
                    c
                    / a
                    * domegadx[jj]
                    * (1 - np.exp(c * (omega[jj] + d - omegaBands[ii][0])))
                )
            else:
                dgdx[ig, :] = (
                    c
                    * domegadx[jj]
                    * (1 - np.exp(c * (omega[jj] + d - omegaBands[ii][0])))
                )
            ig += 1
    if infFilter:
        dgdx = np.clip(dgdx, -infFilter, infFilter)
    return dgdx



"""
GA
GP
ES
PSO

https://deap.readthedocs.io/en/master/tutorials/advanced/constraints.html
https://deap.readthedocs.io/en/master/examples/index.html
"""
import deap


def OptNlOpt(self, x0, xL, xU, SysEq):
    def evalFct(individual):
        """Evaluation function for the individual."""
        x = individual[0]
        return ((x - 5) ** 2 * sin(x) * (x / 3),)

    def feasible(individual):
        """Feasibility function for the individual. Returns True if feasible False
        otherwise."""
        if 3 < individual[0] < 7:
            return True
        return False

    def distance(individual):
        """A distance function to the feasibility region."""
        return (individual[0] - 5.0) ** 2



# -*- coding: utf-8 -*-
"""
Testbook test function for design optimization

Source:  https://dakota.sandia.gov//sites/default/files/docs/6.1/html-ref/textbook.html

Unconstrained:
xOpt = [0.1, 0.1]
fOpt = 0.0

Constrained:
xOpt = [0.5, 0.5]
fOpt = 0.125"""
from DesOptPy import OptimizationProblem
import numpy as np


class Textbook:
    x = [5, 0]

    def SysEq(self):
        self.f = (self.x[0] - 1.0) ** 4 + (self.x[1] - 1.0) ** 4
        self.g1 = self.x[0] ** 2 - self.x[1] / 2.0
        self.g2 = self.x[1] ** 2 - self.x[0] / 2.0


OptTextbook = OptimizationProblem(Textbook)
OptTextbook.Primal = 'SysEq'
OptTextbook.x = 'x'
OptTextbook.xL = [0.5, -2.9]
OptTextbook.xU = [5.8, 2.9]
OptTextbook.x0 = OptTextbook.xU
OptTextbook.f = 'f'
OptTextbook.optimize()
OptTextbook.plotConvergence()

OptTextbook = OptimizationProblem(Textbook)
OptTextbook.Primal = 'SysEq'
OptTextbook.x = 'x'
OptTextbook.xL = [0.5, -2.9]
OptTextbook.xU = [5.8, 2.9]
OptTextbook.x0 = OptTextbook.xU
OptTextbook.f = 'f'
OptTextbook.g = ['g1', 'g2']
OptTextbook.gLimit = [0, 0]
OptTextbook.optimize()
OptTextbook.plotConvergence()
OptTextbook.plotBeforeAfter()



# -*- coding: utf-8 -*-
"""
Rastrigin test function for design optimization

xOpt = 0
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Rastrigin:
    x = np.zeros(
        [
            2,
        ]
    )

    def SysEq(self):
        n = len(self.x)
        self.f = 10 * n + np.sum(
            self.x ** 2 - 10 * np.cos(2.0 * np.pi * self.x)
        )


OptRastrigin = OptimizationProblem(Rastrigin)
OptRastrigin.Primal = 'SysEq'
OptRastrigin.x = 'x'
OptRastrigin.x0 = (
    np.ones(
        2,
    )
    * +2
)
OptRastrigin.xL = (
    np.ones(
        2,
    )
    * -5
)
OptRastrigin.xU = (
    np.ones(
        2,
    )
    * +5
)
OptRastrigin.f = 'f'
OptRastrigin.optimize()
OptRastrigin.plotConvergence()



# -*- coding: utf-8 -*-
"""
Ackley test function for design optimiazation

xOpt = [0.0, 0.0]
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Ackley:
    x = 10

    def SysEq(self):
        a = 20
        b = 0.2
        c = 2 * np.pi
        n = len(self.x)
        s1 = sum(self.x ** 2)
        s2 = sum(np.cos(c * self.x))
        self.f = (
            -a * np.exp(-b * np.sqrt(s1 / n)) - np.exp(s2 / n) + a + np.exp(1)
        )


nx = 10
OptAckley = OptimizationProblem(Ackley)
OptAckley.x = 'x'
OptAckley.x0 = [1] * nx
OptAckley.xL = -5
OptAckley.xU = 5
OptAckley.f = 'f'
OptAckley.Primal = 'SysEq'
OptAckley.optimize()
OptAckley.plotConvergence()



import numpy as np


def printResults(self):
    """


    Returns
    -------
    None.

    """
    lines = '-' * 75
    print()
    # print(lines)
    print('Optimization results - DesOptPy 2.0')
    print()
    # print(lines)
    print(self.ModelName)
    print('Optimization algorithm = ' + self.Alg)
    print()
    print('Optimization model values:')
    # print()
    print('f* = ', end='')
    print(*self.fNormOpt, sep='\n     ', flush=True)
    print()
    if self.gOpt is not None:
        print('g* = ', end='')
        print(*np.array(self.gOpt), sep='\n     ', flush=True)
        print()
    print('x* = ', end='')
    print(*self.xNormOpt[0 : self.nx], sep='\n     ', flush=False)
    print()

    # add function to normalize and denormlaize constraints, change here.
    # print("Evaluation system values")
    # print()
    print('objective response at optimum:')
    for i in range(self.nf):
        print(self.f[i] + ' = ' + str((self.fOpt[i])))
    print()
    if self.g is not None:
        print('constrained response at optimum (limit):')
        # for i in range(self.ng):
        for i, gi in enumerate(self.gOpt):
            if self.gNorm[i]:
                if self.gType[i] == 'upper':
                    r = (gi + 1) * self.gLimit[i]
                elif self.gType[i] == 'lower':
                    r = -(gi - 1) * self.gLimit[i]
            else:
                if self.gType[i] == 'upper':
                    r = gi + self.gLimit[i]
                elif self.gType[i] == 'lower':
                    r = self.gLimit[i] - gi
            if np.size(self.g) == 1 and np.size(r) == 1:
                print(
                    self.g[0]
                    + ' = '
                    + str(r)
                    + ' ('
                    + str(self.gLimit[i])
                    + ', '
                    + self.gType[i]
                    + ')'
                )
            elif np.size(self.g) == 1 and self.ng > 1:
                print(
                    self.g[i]
                    + ' = '
                    + str(r[0])
                    + ' ('
                    + str(self.gLimit[i])
                    + ', '
                    + self.gType[i]
                    + ')'
                )
                for j in range(1, np.size(self.gi)):
                    print(*np.array(self.r[j]), sep='\n     ', flush=True)
            else:
                print(
                    self.g[i]
                    + ' = '
                    + str(r)
                    + ' ('
                    + str(self.gLimit[i])
                    + ', '
                    + self.gType[i]
                    + ')'
                )
        print()
    print('design at optimum (lower bound, upper bound):')
    if self.xVector:
        print(
            self.x
            + ' = '
            + str(self.xOpt[0])
            + ' ('
            + str(self.xL[0])
            + ', '
            + str(self.xU[0])
            + ')'
        )
        for j in range(1, np.size(self.xOpt)):
            print(
                '    '
                + str(self.xOpt[j])
                + ' ('
                + str(self.xL[j])
                + ', '
                + str(self.xU[j])
                + ')'
            )
    else:
        for i in range(self.nx):
            print(
                self.x[i]
                + ' = '
                + str((self.xOpt[i]))
                + ' ('
                + str(self.xL[i])
                + ', '
                + str(self.xU[i])
                + ')'
            )
    print()
    print('t = ' + str(self.tOpt) + ' [h:mm:ss]')
    print()
    if self.nGen:
        print('nGen = ' + str(self.nGen))
    else:
        print('nIt = ' + str(self.nIt))
    print('nEval = ' + str(self.nEval))
    if self.nSensEval is not None:
        print('nSensEval = ' + str(self.nSensEval))
    print()
    if self.RunFolder and self.RemoveRunFolder is False:
        print('See run directory: ' + self.RunDir)
    elif self.RunFolder and self.RemoveRunFolder:
        print('Run cleaned, run directory deleted')
    else:
        print('Local run, all results saved in current directory')
    # print(lines)
    print()



# -*- coding: utf-8 -*-
"""
Schwefel test function for design optimization

xOpt = 420
fOpt = 0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Schwefel:
    def SysEq(self):
        n = len(self.x)
        self.f = 418.9829 * n - np.sum(
            self.x * np.sin(np.sqrt(np.abs(self.x)))
        )


OptSchwefel = OptimizationProblem(Schwefel)
OptSchwefel.Primal = 'SysEq'
OptSchwefel.x = 'x'
nx = 20
OptSchwefel.x0 = np.ones(
    nx,
)
OptSchwefel.xL = (
    np.ones(
        nx,
    )
    * -500
)
OptSchwefel.xU = (
    np.ones(
        nx,
    )
    * +500
)
OptSchwefel.Alg = 'ALHSO'
OptSchwefel.f = 'f'
OptSchwefel.optimize()
OptSchwefel.plotConvergence()



# -*- coding: utf-8 -*-
"""
Styblinski--Tang test function for design optimization

xOpt = -2.9
fOpt = -78.3
"""

from DesOptPy import OptimizationProblem
import numpy as np


class Styblinski:
    def SysEq(self):
        self.f = (
            1.0
            / 2.0
            * (
                self.x ** 4
                - 16.0 * self.x ** 2
                + 5.0 * self.x
                + self.y ** 4
                - 16.0 * self.y ** 2
                + 5.0 * self.y
            )
        )


OptStyblinski = OptimizationProblem(Styblinski)
OptStyblinski.Primal = 'SysEq'
OptStyblinski.x = ['x', 'y']
OptStyblinski.x0 = (
    np.ones(
        2,
    )
    * +0
)
OptStyblinski.xL = (
    np.ones(
        2,
    )
    * -5
)
OptStyblinski.xU = (
    np.ones(
        2,
    )
    * +5
)
OptStyblinski.f = 'f'
OptStyblinski.Alg = 'ALHSO'
OptStyblinski.optimize()
OptStyblinski.plotConvergence()



"""
horrible solution as calling syseq 2x, one obj, one constraint
"""
import numpy as np
from DesOptPy.scaling import normalize, denormalize
import pygmo as pg
import copy


def OptPyGmo(self, x0, xL, xU, SysEq):
    class OptProbPyGMO:
        def fitness(self1, xVal):
            fVal, gVal, fail = SysEq(xVal)
            try:
                gVal = gVal.tolist()
            except:
                pass
            fg = copy.deepcopy(gVal)
            fg.insert(0, fVal)
            self.fAll.append(fVal)
            self.gAll.append(gVal)
            self.xAll.append(xVal)
            return fg

        def get_bounds(self1):
            return (xL, xU)

        def get_nic(self1):
            return self.ng

        def get_nec(self1):
            return 0

        def gradient(self1, xVal):
            return pg.estimate_gradient_h(
                lambda xVal: self1.fitness(xVal), xVal
            )

    """
    PyGMO
    """
    self.gAll = []
    self.fAll = []
    self.xAll = []
    # AlgOptionsChange to AlgOptions
    AlgOptions_nIndiv = 10
    AlgOptions_iter = 10
    AlgOptions_gen = 10
    unconstrained = True
    prob = pg.problem(OptProbPyGMO())
    pop = pg.population(prob=prob, size=AlgOptions_nIndiv)
    prob = pg.problem(OptProbPyGMO())
    # algo = pg.algorithm(uda = pg.nlopt('auglag'))
    # algo.extract(pg.nlopt).local_optimizer = pg.nlopt('var2')
    # pop = pg.population(prob=prob, size=1)

    pop = pg.population(prob=prob, size=AlgOptions_nIndiv)
    if self.Alg[6:] == 'monte_carlo':
        algo = pg.algorithm.monte_carlo(iters=AlgOptions_iter)
    elif self.ng > 0:
        algo = pg.algorithm(
            pg.cstrs_self_adaptive(
                iters=AlgOptions_gen,
                algo=eval(
                    'pg.' + self.Alg[6:] + '(' + str(AlgOptions_nIndiv) + ')'
                ),
            )
        )
        pop.problem.c_tol = [1e-6] * self.ng
    else:
        algo = eval(
            'pg.algorithm(pg.'
            + self.Alg[6:]
            + '('
            + str(AlgOptions_gen)
            + '))'
        )
    pop = algo.evolve(pop)
    xOpt = pop.champion_x
    fOpt = pop.champion_f[0]
    gOpt = pop.champion_f[1:]

    self.xNorm0 = x0
    self.x0 = self.xAll[0]
    self.fNorm0 = self.fAll[0]
    # self.f0 = self.fAll[0]/self.fNormMultiplier
    self.g0 = self.gAll[0]

    self.nIt = None
    self.xIt = None
    self.fIt = None
    self.gIt = None

    if self.g is not None:
        self.gMax = np.max(self.gAll, 1)

    # Denormalization
    self.xOpt = [None] * self.nx
    self.xNormOpt = xOpt
    self.fNormOpt = np.array([fOpt])
    for i in range(self.nx):
        if self.xNorm[i]:
            self.x0[i] = denormalize(x0[i], self.xL[i], self.xU[i])
            self.xOpt[i] = denormalize(xOpt[i], self.xL[i], self.xU[i])
        else:
            self.xOpt[i] = xOpt[i]
    if self.fNorm[0]:
        self.fAll = np.array(self.fAll) * self.f0 / self.fNormMultiplier
        self.fOpt = np.array([fOpt * self.f0 / self.fNormMultiplier])
    else:
        self.fOpt = np.array([fOpt])
    self.gOpt = gOpt



# -*- coding: utf-8 -*-
"""
Title:    RosenbrockSens.py
Units:    -
Author:   E.J. Wehrle
Date:     July 5, 2016
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
Description:

Rosenbrock test function for design optimization

xOpt = [1, 1]
fOpt = 0

-------------------------------------------------------------------------------
"""
from DesOptPy import DesOpt
from DesOptPy import OptAlgOptions
import numpy as np
from scipy.optimize import rosen, rosen_der


def SysEq(x, gc):
    f = rosen(x)
    return f, []


def SensEq(x, f, g, gc):
    dfdx = rosen_der(x)
    return dfdx, []


x0 = (
    np.ones(
        [
            20,
        ]
    )
    * -1
)
xL = (
    np.ones(
        [
            20,
        ]
    )
    * -5
)
xU = (
    np.ones(
        [
            20,
        ]
    )
    * 5
)
gc = []
Alg = 'NLPQLP'
AlgOptions = OptAlgOptions.setDefault(Alg)
AlgOptions.MAXIT = 1000
SensEqList = [SensEq, '', '']
SensCalcList = ['', 'FD', 'AD']
Output = [[]] * len(SensEqList)
Name = ['AnaSens', 'NumSens', 'AutoDiff']
for ii in range(len(SensEqList)):
    xOpt, fOpt, Output[ii] = DesOpt(
        x0=x0,
        xL=xL,
        xU=xU,
        gc=gc,
        SysEq=SysEq,
        SensCalc=SensCalcList[ii],
        SensEq=SensEqList[ii],
        Alg=Alg,
        AlgOptions=AlgOptions,
        DesVarNorm=True,
        deltax=1e-4,
        ResultReport=False,
        StatusReport=False,
        OptNameAdd='RosenSens' + Name[ii],
    )
for ii in range(len(SensEqList)):
    print(Name[ii] + ':')
    print('nIter: ' + str(Output[ii]['nIter']))
    print('nEval: ' + str(Output[ii]['nEval']))
    print('')



# -*- coding: utf-8 -*-
"""
Cylinder head test function for design optimization

Source:  Dakota User's Guide §20.4

xOpt = [2.122, 1.769]
fOpt = -2.461
"""
from DesOptPy import OptimizationProblem
import numpy as np


class CylinderHead:
    x = [1, 1]

    def SysEq(self):
        print(self.x)
        exhaustOffset = 1.34
        exhaustDia = 1.556
        intakeOffset = 3.25
        warranty = 100000 + 15000 * (4 - self.x[1])
        cycleTime = 45 + 4.5 * pow(4 - self.x[1], 1.5)
        wallThickness = (
            intakeOffset - exhaustOffset - (self.x[0] + exhaustDia) / 2
        )
        horsePower = 250.0 + 200.0 * (self.x[0] / 1.833 - 1.0)
        maxStress = 750 + pow(np.fabs(wallThickness), -2.5)
        self.f = -1 * (horsePower / 250 + warranty / 100000)
        self.g1 = maxStress / 1500.0 - 1.0
        self.g2 = 1.0 - warranty / 100000.0
        self.g3 = cycleTime / 60.0 - 1.0


OptCylinderHead = OptimizationProblem(CylinderHead)
OptCylinderHead.Primal = 'SysEq'
OptCylinderHead.x = 'x'
OptCylinderHead.x0 = [1.8, 1.0]
OptCylinderHead.xL = [1.5, 0.0]
OptCylinderHead.xU = [2.164, 4.0]
OptCylinderHead.f = 'f'
OptCylinderHead.g = ['g1', 'g2', 'g3']
OptCylinderHead.gLimit = [0] * 3
OptCylinderHead.Alg = 'SCIPY_SLSQP'
OptCylinderHead.optimize()
OptCylinderHead.plotConvergence()
OptCylinderHead.plotBeforeAfter()



# -*- coding: utf-8 -*-
"""
Himmelblau test function for design optimization

xOpt = [3.0, 2.0]
xOpt = [-2.805118, 3.131312]
xOpt = [-3.779310, -3.283186]
xOpt = [3.584428, -1.848126]
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Himmelblau:
    x = np.zeros(
        [
            2,
        ]
    )

    def SysEq(self):
        self.f = (self.x[0] ** 2 + self.x[1] - 11) ** 2 + (
            self.x[0] + self.x[1] ** 2 - 7
        ) ** 2


OptHimmelblau = OptimizationProblem(Himmelblau)
OptHimmelblau.Primal = 'SysEq'
OptHimmelblau.x = 'x'
OptHimmelblau.x0 = (
    np.ones(
        2,
    )
    * +0
)
OptHimmelblau.xL = (
    np.ones(
        2,
    )
    * -5
)
OptHimmelblau.xU = (
    np.ones(
        2,
    )
    * +5
)
OptHimmelblau.f = 'f'
OptHimmelblau.Alg = 'SOLVOPT'
OptHimmelblau.optimize()
OptHimmelblau.plotConvergence()



# -*- coding: utf-8 -*-
"""
Levy test function for design optimization

xOpt = 1
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Levy:
    x = np.zeros(
        [
            2,
        ]
    )

    def SysEq(self):
        n = len(self.x)
        z = 1.0 + (self.x - 1.0) / 4.0
        self.f = (
            np.sin(np.pi * z[0]) ** 2
            + np.sum(
                (z[:-1] - 1.0) ** 2
                * (1.0 + 10.0 * np.sin(np.pi * z[:-1] + 1.0) ** 2)
            )
            + (z[-1] - 1.0) ** 2 * (1.0 + np.sin(2.0 * np.pi * z[-1]) ** 2)
        )


OptLevy = OptimizationProblem(Levy)
OptLevy.Primal = 'SysEq'
OptLevy.x = 'x'
OptLevy.x0 = (
    np.ones(
        2,
    )
    * +0
)
OptLevy.xL = (
    np.ones(
        2,
    )
    * -5
)
OptLevy.xU = (
    np.ones(
        2,
    )
    * +5
)
OptLevy.f = 'f'
OptLevy.optimize()
OptLevy.plotConvergence()



# -*- coding: utf-8 -*-
"""
Test function for design optimization

f* = -3456.
x* = 24.0, 12.0, 12.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class TP037:
    x = [1, 1]

    def SysEq(self):
        print(self.x)
        self.f = -self.x[0] * self.x[1] * self.x[2]
        self.g = (
            np.ones(
                [
                    2,
                ]
            )
            * 0.0
        )
        self.g[0] = self.x[0] + 2.0 * self.x[1] + 2.0 * self.x[2] - 72.0
        self.g[1] = -self.x[0] - 2.0 * self.x[1] - 2.0 * self.x[2]


OptTP037 = OptimizationProblem(TP037)
OptTP037.Primal = 'SysEq'
OptTP037.x = 'x'
OptTP037.x0 = [20] * 3
OptTP037.xL = [0] * 3
OptTP037.xU = [40] * 3
OptTP037.f = 'f'
OptTP037.g = ['g']
OptTP037.gLimit = [0] * 2
OptTP037.optimize()
OptTP037.plotConvergence()
OptTP037.plotBeforeAfter()



import copy
import numpy as np
from DesOptPy.scaling import normalize, denormalize


def readHistory(self):
    # make function of new file
    import pyOpt

    # read values from history
    OptHist = pyOpt.History(self.Name, 'r')
    xAll = OptHist.read([0, -1], ['x'])[0]['x']
    fAll = OptHist.read([0, -1], ['obj'])[0]['obj']
    gAll = OptHist.read([0, -1], ['con'])[0]['con']
    gNablaIt = OptHist.read([0, -1], ['grad_con'])[0]['grad_con']
    fNablaIt = OptHist.read([0, -1], ['grad_obj'])[0]['grad_obj']
    failIt = OptHist.read([0, -1], ['fail'])[0]['fail']
    OptHist.close()

    # denormalize design variable values
    xNormAll = copy.deepcopy(xAll)
    for ei in range(len(xAll)):
        for xi, xNormi in enumerate(self.xNorm):
            if xNormi:
                xAll[ei][xi] = denormalize(
                    xAll[ei][xi], self.xL[xi], self.xU[xi]
                )
    # NLPQLP g>0
    if self.Alg == 'NLPQLP':
        gAll = [g * -1 for g in gAll]

    # convert to list
    # TODO should be np.arrays??!!
    xAll = [i.tolist() for i in xAll]
    xNormAll = [i.tolist() for i in xNormAll]
    fAll = [i.tolist() for i in fAll]
    gAll = [i.tolist() for i in gAll]
    fNablaIt = [i.tolist() for i in fNablaIt]
    gNablaIt = [i.tolist() for i in gNablaIt]
    failIt = [i.tolist() for i in failIt]

    # initialize assignment to object
    self.nIt = len(fNablaIt)
    self.fIt = [None] * self.nIt
    self.fNormIt = [None] * self.nIt
    self.xIt = [None] * self.nIt
    self.xNormIt = [None] * self.nIt
    self.gIt = [None] * self.nIt
    self.gMaxIt = [None] * self.nIt

    self.gNablaIt = gNablaIt
    self.fNablaIt = fNablaIt

    # #TODO remove this
    # len(fAll) not = nEval for gradientbased!!!
    # if self.nEval == len(fAll):
    #     pass
    # else:
    #     raise Exception("something evil")

    # check if non-iteration-based algorithm
    if self.nIt == 0 and self.nEval > 0:
        self.nIt = None
        self.xNorm0 = xNormAll[0]
        if self.g is not None:
            self.g0 = np.array(gAll[0])
            self.gOpt = np.array(gAll[-1])
    else:

        for ii in range(self.nIt):
            Posdg = OptHist.cues['grad_con'][ii][0]
            Posf = OptHist.cues['obj'][ii][0]
            iii = 0
            while Posdg > Posf:
                iii = iii + 1
                try:
                    Posf = OptHist.cues['obj'][iii][0]
                except:
                    Posf = Posdg + 1
            iii = iii - 1
            if self.fNorm[0]:
                if self.f0 == 0:
                    self.fIt[ii] = (
                        np.array(fAll[iii])
                        / self.fNormMultiplier
                        * self.fMinMax
                    )
                else:
                    self.fIt[ii] = (
                        np.array(fAll[iii])
                        * abs(self.f0)
                        / self.fNormMultiplier
                        * self.fMinMax
                    )
            else:
                self.fIt[ii] = fAll[iii]
            self.fNormIt[ii] = fAll[iii]

            self.xIt[ii] = xAll[iii][0 : self.nx]
            self.xNormIt[ii] = xNormAll[iii][0 : self.nx]
            if self.g is not None:
                self.gIt[ii] = gAll[iii]
                self.gMaxIt[ii] = np.max(gAll[iii])
        self.xNorm0 = xNormAll[0][0 : self.nx]
        if self.g is not None:
            self.gOpt = np.array(self.gIt[-1])
            self.g0 = np.array(self.gIt[0])

        self.fNablaIt = np.array(self.fNablaIt) * self.fMinMax
        self.fNormNablaIt = copy.deepcopy(self.fNablaIt)
        self.fNablaOpt = copy.deepcopy(self.fNablaIt[-1])
        self.fNormNablaOpt = copy.deepcopy(self.fNablaIt[-1])

        if self.fNorm[0]:
            if self.f0 == 0:
                self.fNablaIt = self.fNablaIt / self.fNormMultiplier
                self.fNablaOpt = self.fNablaOpt / self.fNormMultiplier
            else:
                self.fNablaIt = (
                    self.fNablaIt / self.fNormMultiplier * abs(self.f0)
                )
                self.fNablaOpt = (
                    self.fNablaOpt / self.fNormMultiplier * abs(self.f0)
                )
        self.gNablaOpt = np.array(self.gNablaIt[-1]).reshape(self.ng, self.nx)
    self.xAll = xAll
    self.xNormAll = xNormAll * self.fMinMax
    self.fAll = fAll * self.fMinMax
    self.gAll = gAll
    if self.g is not None:
        self.gMax = np.max(self.gAll, 1)



"""
horrible solution as calling syseq 2x, one obj, one constraint

def BOBYQA(func, x0, args, tol=1e-6):
    # Set optimizer
    opt = nlopt.opt(nlopt.LN_BOBYQA, len(x0))
    res = results()
    prob = opt_problem(func, res, args)
    opt.set_ftol_abs(tol)
    opt.set_min_objective(prob.objfun)
    opt.add_inequality_constraint(prob.constrfun)
    res.x = opt.optimize(x0)
    res.fun = opt.last_optimum_value()
    return res


class results:
    def __init__(self):
        self.x = None
        self.fun = None
        self.nit = 0
        self.nfev = 0


class opt_problem:
    def __init__(self, func, res, args):
        self.func = func
        self.curr_x = None
        self.fobj = 0.0
        self.grad = None
        self.res = res
        self.args = args

    def objfun(self, x, *_):
        if self.curr_x != x:
            self.res.nit += 1
            self.res.nfev += 1
            self.curr_x = x
            self.res.x = x
            self.fobj, self.grad = self.func(x, *self.args)
        return self.fobj

    def constrfun(self, x, *_):
        if self.curr_x != x:
            self.res.nit += 1
            self.res.nfev += 1
            self.curr_x = x
            self.res.x = x
            self.fobj, self.grad = self.func(x, *self.args)
        return self.grad
"""
import numpy as np
from DesOptPy.scaling import normalize, denormalize
import nlopt


def OptNlOpt(self, x0, xL, xU, SysEq):
    def ObjFn(self, xVal, grad):
        if np.array_equal(self.xLast, xVal) != True:
            self.fVal, self.gVal, flag = SysEq(xVal)
            self.xLast = xVal.copy()
            self.fAll.append(self.fVal)
            self.gAll.append(self.gVal)
            self.xAll.append(xVal)
            print(self.fVal)
        return float(self.fVal)

    def ConFn(self, xVal, grad):
        if np.array_equal(self.xLast, xVal) != True:
            self.fVal, self.gVal, flag = SysEq(xVal)
            self.xLast = xVal.copy()
            self.fAll.append(self.fVal)
            self.gAll.append(self.gVal)
            self.xAll.append(xVal)
        return self.gVal

    self.gAll = []
    self.fAll = []
    self.xAll = []

    algorithm = nlopt.LD_MMA
    algorithm = nlopt.LN_COBYLA
    # algorithm = nlopt.LN_COBYLA
    # algorithm = nlopt.LN_BOBYQA

    opt = nlopt.opt(algorithm, self.nx)
    # opt = nlopt.opt(algorithm, self.nx)
    opt.set_min_objective(ObjFn)
    # opt.set_max_objective(f)
    opt.set_lower_bounds(self.xL)
    opt.set_upper_bounds(self.xU)
    opt.add_inequality_constraint(ConFn)
    # opt.add_inequality_constraint(lambda x, grad: ConFn(x, grad))

    # opt.set_stopval(stopval)
    # opt.set_ftol_rel(tol)
    # opt.set_ftol_rel(tol)
    # opt.set_xtol_abs(tol)
    # opt.set_xtol_rel(tol)
    # opt.set_maxeval(maxeval)
    # opt.set_maxtime(maxtime)
    xopt = opt.optimize(self.x0)

    xOpt = np.array(Results.x)
    fOpt = np.array([Results.fun])

    self.xNorm0 = x0
    self.x0 = self.xAll[0]
    self.f0 = self.fAll[0]
    self.g0 = self.gAll[0]
    self.fNablaOpt = Results.jac
    self.nIt = Results.nit
    self.xIt = None
    self.fIt = None
    self.gIt = None

    # Todo this is ugly
    self.nIt = Results.nfev
    try:
        self.nSensEval = Results.njev
    except:
        self.nSensEval = None
    self.inform = Results.success

    if 'SLSQP' in (self.Alg).upper():
        self.fNablaOpt = Results.jac
    elif 'trust-constr' in (self.Alg).lower():
        self.fNablaOpt = Results.grad
        self.gNablaOpt = Results.jac[0]

    if self.g is not None:
        self.gMax = np.max(self.gAll, 1)

    # Denormalization
    self.xOpt = [None] * self.nx
    self.xNormOpt = xOpt
    self.fNormOpt = fOpt
    for i in range(self.nx):
        if self.xNorm[i]:
            self.xOpt[i] = denormalize(xOpt[i], self.xL[i], self.xU[i])
        else:
            self.xOpt[i] = xOpt[i]
    if self.fNorm[0]:
        self.fOpt = fOpt * self.f0 / self.fNormMultiplier
    else:
        self.fOpt = fOpt
    self.gOpt = self.gVal



# -*- coding: utf-8 -*-
"""
Cylinder head test function for design optimization

Source:  Dakota User's Guide §20.4

xOpt = [2.122, 1.769]
fOpt = -2.461
"""
from DesOptPy import OptimizationProblem
import numpy as np


class CylinderHead:
    x = 1
    y = 2

    def SysEq(self):
        exhaustOffset = 1.34
        exhaustDia = 1.556
        intakeOffset = 3.25
        self.warranty = 100000 + 15000 * (4 - self.y)
        self.cycleTime = 45 + 4.5 * pow(4 - self.y, 1.5)
        wallThickness = (
            intakeOffset - exhaustOffset - (self.x + exhaustDia) / 2
        )
        horsePower = 250.0 + 200.0 * (self.x / 1.833 - 1.0)
        self.maxStress = 750 + pow(np.fabs(wallThickness), -2.5)
        self.f = -1 * (horsePower / 250 + self.warranty / 100000)


OptCylinderHead = OptimizationProblem(CylinderHead)
OptCylinderHead.Primal = 'SysEq'
OptCylinderHead.x = ['x', 'y']
OptCylinderHead.xDelta = 1e-6
OptCylinderHead.x0 = [1.8, 1.0]
OptCylinderHead.xL = [1.5, 0.0]
OptCylinderHead.xU = [2.164, 4.0]
OptCylinderHead.f = 'f'
OptCylinderHead.g = ['maxStress', 'warranty', 'cycleTime']
OptCylinderHead.gType = ['upper', 'lower', 'upper']
OptCylinderHead.gLimit = [1500, 100000, 60]
OptCylinderHead.optimize()
OptCylinderHead.plotConvergence()
OptCylinderHead.plotBeforeAfter()



from packaging import version
import os
import numpy as np


def checkProblem(self):
    # idiot check
    if self.x == None:
        raise Exception('Design variables x not set. ')
    # if self.xL == None:
    #    raise Exception("Lower bound of design variables xL not set. ")
    # if self.xU == None:
    #    raise Exception("Lower bound of design variables xU not set. ")
    if self.f == None:
        raise Exception('Objective f not set. ')
    if self.Primal == None:
        raise Exception('Primal not set. ')


def checkAlgorithms():
    """


    Returns
    -------
    None.

    """
    print('The following algorithms are available')
    try:
        import pyOpt

        pyOptAlgList = [
            'ALGENCAN',
            'ALHSO',
            'ALPSO',
            'COBYLA',
            'CONMIN',
            'FILTERSD',
            'FSQP',
            'GCMMA',
            'IPOPT',
            'KSOPT',
            'MIDACO',
            'MMA',
            'MMFD',
            'NLPQL',
            'NLPQLP',
            'NSGA2',
            'PSQP',
            'SDPEN',
            'SLSQP',
            'SNOT',
            'SOLVOPT',
        ]
        algAvail = '\npyOpt algorithms available:'
        algNotAvail = '\npyOpt algorithms NOT AVAILABLE:'
        na = 0
        for i in pyOptAlgList:
            try:
                eval('pyOpt.' + i + '()')
                algAvail += '\n' + i + ' available'
            except:
                algNotAvail += '\n' + i + ' NOT AVAILABLE'
                na = 1
        print('\033[32m' + algAvail + '\033[0m')
        if na:
            print('\033[31m' + algNotAvail + '\033[0m')
    except:
        print('\033[41m' + 'pyOpt NOT INSTALLED' + '\033[0m')

    try:
        import scipy
        import scipy.optimize as spopt

        SciPyAlgList = ['slsqp', 'trust-constr', 'differential_evolution']
        algAvail = '\nSciPy ' + scipy.__version__ + ' algorithms available:'
        algNotAvail = '\nSciPy algorithms NOT AVAILABLE:'
        algAvail += '\n' + 'SLSQP' + ' available'
        algAvail += '\n' + 'trust-constr' + ' available'
        na = 0
        if version.parse(scipy.__version__) >= version.parse('1.7.0'):
            for i in SciPyAlgList[2:]:
                try:
                    eval('spopt.' + i)
                    algAvail += '\n' + i + ' available'
                except:
                    algNotAvail += '\n' + i + ' NOT AVAILABLE'
                    na = 1
            print('\033[32m' + algAvail + '\033[0m')
            if na:
                print('\033[31m' + algNotAvail + '\033[0m')
        else:
            print('\nDesOptPy tested with version SciPy version 1.7.0')
            print('update to this version, e.g.')
            print('    pip install scipy -U')
    except:
        print('\033[31m' + 'SciPy NOT INSTALLED' + '\033[0m')

    try:
        import pygmo

        algAvail = 'pyGMO algorithms available'
        print('\033[32m' + algAvail + '\033[0m')
    except:
        print('\033[31m' + 'pyGMO NOT INSTALLED' + '\033[0m')



from bokeh import palettes
from bokeh.plotting import figure, show, output_file, save
from bokeh.layouts import column, layout
from bokeh.models import LinearAxis, Range1d, Legend, ColumnDataSource
from bokeh.models.tools import BoxZoomTool, PanTool, ResetTool, HoverTool
import numpy as np
import panel as pn
import os


# TODO write out details at beginning including xL, xU, nx etc. Or just entire model self

"""
https://github.com/holoviz/panel/issues/1496

"""

# ds = ColumnDataSource({"x": [0], "fAll": [0], "gMax": [0]})

showPlot = True
savePlot = False
OptName = os.getcwd().split(os.sep)[-1][6:]
Dir = os.getcwd() + os.sep + os.getcwd().split(os.sep)[-1][6:]

font = 'palantino'
fontSize = '12pt'


def colorPalette(nColors):
    # Palettes: https://docs.bokeh.org/en/latest/docs/reference/palettes.html
    if nx < 11:
        colorsx = palettes.d3['Category10'][10]
    elif nx < 21:
        colorsx = (
            palettes.d3['Category20'][20][::2]
            + palettes.d3['Category20'][20][1::2]
        )
    else:
        colorsx = (
            palettes.d3['Category20'][20][::2]
            + palettes.d3['Category20'][20][1::2]
            + palettes.d3['Category20c'][20][::4]
            + palettes.d3['Category20b'][20][::4]
            + palettes.d3['Category20c'][20][1::4]
            + palettes.d3['Category20b'][20][1::4]
            + palettes.d3['Category20c'][20][2::4]
            + palettes.d3['Category20b'][20][2::4]
            + palettes.d3['Category20c'][20][3::4]
            + palettes.d3['Category20b'][20][3::4]
        )
    if nx > 60:
        colorsx *= nx // 60 + 1
    return colorsx[:nColors]


def readHistory():
    import pyOpt

    OptHist = pyOpt.History(Dir, 'r')
    xAll = OptHist.read([0, -1], ['x'])[0]['x']
    fAll = OptHist.read([0, -1], ['obj'])[0]['obj']
    gAll = OptHist.read([0, -1], ['con'])[0]['con']
    gMax = np.max(gAll, 1)
    return (xAll, fAll, gAll, gMax)


def plotStyle(p, axes, colors):
    # colors
    p.xgrid.visible = False
    p.ygrid.visible = False
    p.background_fill_alpha = 0.0
    p.border_fill_alpha = 0.0
    p.outline_line_alpha = 0.0

    # title
    p.title.align = 'left'
    p.title.text_font = font
    p.title.text_font_size = '14pt'
    p.title.text_color = 'black'
    p.title.text_font_style = 'normal'

    # x axis
    p.xaxis.axis_label_text_font = font
    p.xaxis.axis_label_text_font_style = 'normal'
    p.xaxis.axis_label_text_font_size = fontSize
    p.xaxis.axis_label_text_color = 'black'
    p.xaxis.major_label_text_font = font
    p.xaxis.major_label_text_font_style = 'normal'
    p.xaxis.major_label_text_font_size = fontSize
    p.xaxis.major_label_text_color = 'black'
    p.xaxis.major_tick_in = 5
    p.xaxis.major_tick_out = 0
    p.xaxis.minor_tick_in = 0  # 2
    p.xaxis.minor_tick_out = 0

    # y axis
    # p.yaxis.axis_label_text_align = "horizontal"
    p.yaxis.axis_label_text_font = font
    p.yaxis.axis_label_text_font_style = 'normal'
    p.yaxis.axis_label_text_font_size = fontSize
    p.yaxis.major_label_text_font = font
    p.yaxis.major_label_text_font_style = 'normal'
    p.yaxis.major_label_text_font_size = fontSize
    if axes == 2:
        p.yaxis[0].axis_label_text_color = colors[0]
        p.yaxis[1].axis_label_text_color = colors[3]
        p.yaxis[0].major_label_text_color = colors[0]
        p.yaxis[1].major_label_text_color = colors[3]
        p.yaxis[0].major_tick_line_color = colors[0]
        p.yaxis[1].major_tick_line_color = colors[3]
        p.yaxis[0].axis_line_color = colors[0]
        p.yaxis[1].axis_line_color = colors[3]
    p.yaxis.major_tick_in = 5
    p.yaxis.major_tick_out = 0
    p.yaxis.minor_tick_in = 0  # 2
    p.yaxis.minor_tick_out = 0

    # toolbar
    p.toolbar.logo = None
    # p.add_tools(PanTool(dimensions="width"))
    # p.add_tools(HoverTool(
    #     tooltips=[
    #         ('evaluation',   '@x{0}'            ),
    #         ('objective',  "@y2{0.009}"),
    #         ('max constraint', '@y3{0.000}'),
    #     ],
    #     line_policy='nearest',
    #     mode='vline'))

    return p


def update(event=None):
    global i
    xAll, fAll, gAll, gMax = readHistory()
    it = np.array(range(len(fAll)))
    if i < len(it):
        i += 1
        ds_fg.stream(
            {'x': [it[-1]], 'fAll': [fAll[-1][0]], 'gMax': [gMax[-1]]}
        )
        # ds_x.stream({"x": [it[-1]], "xAll": [xAll[-1].tolist()]})

    for i in range(nx):
        ds_x[i].stream(
            {
                'x': [it[-1]],
                # "xAll": np.array(xAlcolorPalette(4)l).reshape(nx, len(it)).tolist(),
                'xAll': [xAll[-1][i]],
            }
        )

    for i in range(ng):
        ds_g[i].stream(
            {
                'x': [it[-1]],
                'gAll': [gAll[-1][i]],
            }
        )


# Read in model information
import pickle

with open(OptName + '.pkl', 'rb') as f:
    ModelInfo = pickle.load(f)


# initial
xAll, fAll, gAll, gMax = readHistory()
i = len(fAll)
it = np.array(range(len(fAll)))

# data as data stream
# ds = ColumnDataSource({"x": it, "y": fAll, "gMax":gMax})
ds_fg = ColumnDataSource(
    {
        'x': it.tolist(),
        'fAll': np.array(fAll).reshape(len(it)).tolist(),
        'gMax': gMax.tolist(),
    }
)
nx = len(xAll[0])


# initialize objective and constraint convergence
p_fg = figure(
    sizing_mode='stretch_both', toolbar_location='above', toolbar_sticky=False
)
p_fg.xaxis.axis_label = 'evaluation'

# second y axis
gDelta = (max(gMax) - min(gMax)) * 0.025
p_fg.extra_y_ranges = {
    'constraint': Range1d(
        start=min(gMax) - gDelta,  # np.sign(max(gMax)) * gDelta,
        end=max(gMax) + gDelta,  # np.sign(min(gMax)) * gDelta,
    )
}
p_fg.add_layout(LinearAxis(y_range_name='constraint'), 'right')
fDelta = (max(fAll)[0] - min(fAll)[0]) * 0.025
p_fg.y_range = Range1d(
    start=min(fAll)[0] - fDelta,  # np.sign(max(fAll)[0]) * fDelta,
    end=max(fAll)[0] + fDelta,  # np.sign(min(fAll)[0]) * fDelta,
)
# p_fg.y_range.range_padding = fDelta[0]*0.01
# p_fg.y_range.range_padding = fDelta[0]*0.01

colors_fg = colorPalette(10)
# plot style
p_fg = plotStyle(p_fg, 2, colors_fg)
p_fg.yaxis[0].axis_label = 'objective value'
p_fg.yaxis[1].axis_label = 'maximum constraint value'


# plot
p_fg.line(x='x', y='fAll', line_color=colors_fg[0], source=ds_fg)
p_fg.circle(
    x='x',
    y='fAll',
    fill_color='white',
    line_color='white',
    source=ds_fg,
    size=10,
)
p_fg.circle(
    x='x',
    y='fAll',
    fill_color=colors_fg[0],
    line_color=colors_fg[0],
    source=ds_fg,
    size=5,
)
p_fg.line(
    x='x',
    y='gMax',
    line_color=colors_fg[3],
    y_range_name='constraint',
    source=ds_fg,
)
p_fg.circle(
    x='x',
    y='gMax',
    fill_color='white',
    line_color='white',
    y_range_name='constraint',
    source=ds_fg,
    size=10,
)
p_fg.circle(
    x='x',
    y='gMax',
    fill_color=colors_fg[3],
    line_color=colors_fg[3],
    y_range_name='constraint',
    source=ds_fg,
)


# Design variable convergence
"""
https://medium.com/@andrewm4894/bokeh-battles-part-1-multi-line-plots-311109992fdc
https://towardsdatascience.com/draw-beautiful-and-interactive-line-charts-using-bokeh-in-python-9f3e11e0a16e
"""


# initialize design variable convergence
p_x = figure(
    sizing_mode='stretch_both', toolbar_location='above', toolbar_sticky=False
)
p_x.xaxis.axis_label = 'evaluation'
p_x.yaxis.axis_label = 'normalized design variable value'
colorsx = colorPalette(nx)
# plot style
p_x = plotStyle(p_x, 1, colorsx)

# plot
# p_x.multi_line(
#     xs="x",
#     ys="xAll",
#     source=ds_x,
#     color=["red", "blue"],
# )
ds_x = [[]] * nx
for i in range(nx):
    ds_x[i] = ColumnDataSource(
        {
            'x': it.tolist(),
            # "xAll": np.array(xAll).reshape(nx, len(it)).tolist(),
            'xAll': np.array(xAll)[:, i].tolist(),
        }
    )

    p_x.line(
        x='x',
        y='xAll',
        source=ds_x[i],
        color=colorsx[i],
        legend_label='design variable ' + str(i + 1),
    )  # legend_label=r"$\\hat{x}_"+str(i+1)+"$")
    p_x.circle(
        x='x',
        y='xAll',
        source=ds_x[i],
        legend_label='design variable ' + str(i + 1),
        fill_color='white',
        line_color='white',
        size=10,
    )
    p_x.circle(
        x='x',
        y='xAll',
        source=ds_x[i],
        color=colorsx[i],
        size=5,
        legend_label='design variable ' + str(i + 1),
    )

p_x.add_layout(p_x.legend[0], 'right')
p_x.legend.click_policy = 'hide'
p_x.legend.background_fill_alpha = 0.0
p_x.legend.border_line_alpha = 0.0
p_x.legend.label_text_font = font
p_x.legend.label_text_font_size = '12pt'


# initialize design variable convergence
ng = len(gAll[0])
p_g = figure(
    sizing_mode='stretch_both', toolbar_location='above', toolbar_sticky=False
)
p_g.xaxis.axis_label = 'evaluation'
p_g.yaxis.axis_label = 'constraint value'
colors_g = colorPalette(ng)
# plot style
p_g = plotStyle(p_g, 1, colors_g)

# plot
# p_x.multi_line(
#     xs="x",
#     ys="xAll",
#     source=ds_x,
#     color=["red", "blue"],
# )
ds_g = [[]] * ng
for i in range(ng):
    ds_g[i] = ColumnDataSource(
        {
            'x': it.tolist(),
            # "xAll": np.array(xAll).reshape(nx, len(it)).tolist(),
            'gAll': np.array(gAll)[:, i].tolist(),
        }
    )

    p_g.line(
        x='x',
        y='gAll',
        source=ds_g[i],
        color=colors_g[i],
        legend_label='constraint ' + str(i + 1),
    )  # legend_label=r"$\\hat{x}_"+str(i+1)+"$")
    p_g.circle(
        x='x',
        y='gAll',
        source=ds_g[i],
        legend_label='constraint ' + str(i + 1),
        fill_color='white',
        line_color='white',
        size=10,
    )
    p_g.circle(
        x='x',
        y='gAll',
        legend_label='constraint ' + str(i + 1),
        source=ds_g[i],
        color=colors_g[i],
        size=5,
    )

p_g.add_layout(p_g.legend[0], 'right')
p_g.legend.click_policy = 'hide'
p_g.legend.background_fill_alpha = 0.0
p_g.legend.border_line_alpha = 0.0
p_g.legend.label_text_font = font
p_g.legend.label_text_font_size = '12pt'


# layout
"""
https://docs.bokeh.org/en/latest/docs/user_guide/layout.html
"""
plotLayout = layout(
    [
        [p_fg],
        [p_x],
        [p_g],
    ],
    sizing_mode='stretch_both',
)

pane = pn.panel(plotLayout).servable()
pn.state.add_periodic_callback(update, 100)

if showPlot:
    show(plotLayout)

if savePlot:
    output_file(filename='Try.html', title='DesOptPy')
    save(plotLayout)



# -*- coding: utf-8 -*-
"""
Eggholder test function for design optimization

xOpt = []
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Eggholder:
    z = [1, 1]

    def SysEq(self):
        x = self.z[0]
        y = self.z[1]
        self.f = -(y + 47.0) * np.sin(
            np.sqrt(abs(y + x / 2.0 + 47.0))
        ) - x * np.sin(np.sqrt(abs(x - (y + 47.0))))


OptEggholder = OptimizationProblem(Eggholder)
OptEggholder.x = 'z'
OptEggholder.x0 = [0] * 2
OptEggholder.xL = -5
OptEggholder.xU = 5
OptEggholder.f = 'f'
OptEggholder.fNorm = False
OptEggholder.Alg = 'SLSQP'
OptEggholder.Primal = 'SysEq'
OptEggholder.optimize()
OptEggholder.plotConvergence()



import datetime
import os
import copy
import shutil
import numpy as np

try:
    import cpuinfo
except:
    pass
from DesOptPy.scaling import normalize, denormalize
from DesOptPy.tools import printResults, checkProblem
from DesOptPy import plotting

# from DesOptPy import plotting


def OptimizationProblem(Model):
    class Opt(Model):
        ModelName = str(Model.__name__)
        fType = 'min'
        x = None
        f = None
        g = None
        xL = None
        xU = None

        Primal = None
        Sensitivity = None

        ng = None
        nf = None
        nx = None

        xOpt = None
        fOpt = None
        gOpt = None

        fNabla = None
        gNabla = None

        fNorm = True
        xNorm = True
        gNorm = True

        gType = None

        f0 = None
        g0 = None

        xLast = None

        gLimit = [None]

        nGen = None
        nSensEval = None

        fNormMultiplier = 1e3
        xDelta = 1e-4
        Alg = 'SLSQP'

        pyOptAlg = None
        SciPyAlg = None
        pyGmoAlg = None
        nloptAlg = None

        PrintOutput = True
        Alarm = False
        RunFolder = True
        RemoveRunFolder = True
        SaveEvaluations = False
        OS = os.uname()[0]
        Computer = os.uname()[1]
        architecture = os.uname()[4]
        nProcessors = os.cpu_count()
        userName = os.getlogin()
        Monitoring = True
        RunDir = None
        MainDir = None
        try:
            cpu = cpuinfo.get_cpu_info()['brand']
        except:
            cpu = None

        # class KaruschKuhnTucker:
        from DesOptPy.plotting import plotConvergence
        from DesOptPy.postprocessingNumerical import checkKKT, calcShadowPrices

        def readHistory(self):
            # make function of new file
            import pyOpt

            OptHist = pyOpt.History(self.Name, 'r')
            xAll = OptHist.read([0, -1], ['x'])[0]['x']
            xNormAll = copy.deepcopy(xAll)
            # xNormAll = xAll.copy()
            for ei in range(len(xAll)):
                for xi, xNormi in enumerate(self.xNorm):
                    if xNormi:
                        xAll[ei][xi] = denormalize(
                            xAll[ei][xi], self.xL[xi], self.xU[xi]
                        )
            fAll = OptHist.read([0, -1], ['obj'])[0]['obj']
            gAll = OptHist.read([0, -1], ['con'])[0]['con']
            gNablaIt = OptHist.read([0, -1], ['grad_con'])[0]['grad_con']
            if self.Alg == 'NLPQLP':
                gAll = [g * -1 for g in gAll]
            fNablaIt = OptHist.read([0, -1], ['grad_obj'])[0]['grad_obj']
            failIt = OptHist.read([0, -1], ['fail'])[0]['fail']
            OptHist.close()
            xAll = [i.tolist() for i in xAll]
            xNormAll = [i.tolist() for i in xNormAll]
            fAll = [i.tolist() for i in fAll]
            gAll = [i.tolist() for i in gAll]
            fNablaIt = [i.tolist() for i in fNablaIt]
            gNablaIt = [i.tolist() for i in gNablaIt]
            failIt = [i.tolist() for i in failIt]

            self.nIt = len(fNablaIt)
            self.fIt = [None] * self.nIt
            self.xIt = [None] * self.nIt
            self.xNormIt = [None] * self.nIt
            self.gIt = [None] * self.nIt
            self.gMaxIt = [None] * self.nIt
            for ii in range(self.nIt):
                Posdg = OptHist.cues['grad_con'][ii][0]
                Posf = OptHist.cues['obj'][ii][0]
                iii = 0
                while Posdg > Posf:
                    iii = iii + 1
                    try:
                        Posf = OptHist.cues['obj'][iii][0]
                    except:
                        Posf = Posdg + 1
                iii = iii - 1
                self.fIt[ii] = fAll[iii]
                self.xIt[ii] = xAll[iii][0 : self.nx]
                self.xNormIt[ii] = xNormAll[iii][0 : self.nx]
                if self.g is not None:
                    self.gIt[ii] = gAll[iii]
                    self.gMaxIt[ii] = np.max(gAll[iii])

            self.xAll = xAll
            self.xNormAll = xNormAll
            self.fAll = fAll
            self.gAll = gAll
            self.gNablaIt = gNablaIt
            self.fNablaIt = fNablaIt
            if self.g is not None:
                self.gOpt = self.gIt[-1]
                self.gMax = np.max(self.gAll, 1)
                self.g0 = self.gIt[0]

        def optimize(self):

            self.Model = Model
            # self.ModelName = str(self.Model.__name__)
            self.t0 = datetime.datetime.now()
            self.t0str = self.t0.strftime('%Y%m%d%H%M%S')
            self.Name = self.ModelName + self.Alg + self.t0str
            self.nEval = 0

            checkProblem(self)
            if type(self.f) == str:
                self.f = [self.f]

            # set general sizes of optimization problem
            self.nx = max(np.size(self.x), np.size(self.x0), np.size(self.xL))
            self.nf = np.size(self.f)
            if self.g == None or self.g == []:
                self.ng = 0
            else:
                self.ng = max(np.size(self.g), np.size(self.gLimit))

            # check if one parameter for all design variables, i.e. vector
            if np.size(self.x) == 1 and self.nx > 1:
                self.xVector = True
            else:
                self.xVector = False

            # # check if one parameter for all constraints, i.e. vector
            if np.size(self.g) == 1 and self.ng > 1:
                self.gVector = True
            else:
                self.gVector = False

            # Reformat x0, xL and xU to be np.arrays and of proper size
            if type(self.x0) == list:
                self.x0 = np.array(self.x0)
            if type(self.xL) == list:
                self.xL = np.array(self.xL)
            if type(self.xU) == list:
                self.xU = np.array(self.xU)
            if np.size(self.xL) == 1 and self.nx > 1:
                self.xL = np.array([self.xL] * self.nx)
                self.xU = np.array([self.xU] * self.nx)

            # Reformat and set normalization (scaling)
            if (
                self.xNorm == None
                or self.xNorm == True
                or self.xNorm == [True]
            ):
                self.xNorm = [True] * self.nx
            elif self.xNorm == False or self.xNorm == [False]:
                self.xNorm = [False] * self.nx

            if (
                self.fNorm == None
                or self.fNorm == True
                or self.fNorm == [True]
            ):
                self.fNorm = [True] * self.nf
            elif self.fNorm == False or self.fNorm == [False]:
                self.fNorm = [False] * self.nf

            if (
                self.gNorm == None
                or self.gNorm == True
                or self.gNorm == [True]
                or self.gNorm == []
            ):
                self.gNorm = [True] * self.ng
            elif self.gNorm == False or self.gNorm == [False]:
                self.gNorm = [False] * self.ng
            for i, gLimiti in enumerate(self.gLimit):
                if gLimiti == 0:
                    self.gNorm[i] = False

            if (
                self.gType == None
                or self.gType == 'upper'
                or self.gType == ['upper']
            ):
                self.gType = ['upper'] * self.ng
            elif self.gType == 'lower' or self.gType == ['lower']:
                self.xNorm = ['lower'] * self.nx
            # TODO equality constraints?

            # TODO
            # if self.fType.upper()[0:3] == "MIN":
            #
            #

            # Optimization Algorithm
            self.SciPyAlg = False
            self.pyOptAlg = False
            self.pyGmoAlg = False
            self.nloptAlg = False
            if (self.Alg).upper() in {
                'ALGENCAN',
                'ALHSO',
                'ALPSO',
                'COBYLA',
                'CONMIN',
                'FILTERSD',
                'FSQP',
                'GCMMA',
                'IPOPT',
                'KSOPT',
                'MIDACO',
                'MMA',
                'MMFD',
                'NLPQLP',
                'NSGA2',
                'PSQP',
                'SDPEN',
                'SLSQP',
                'SOLVOPT',
            }:
                self.pyOptAlg = True
            elif (self.Alg[:5]).upper() == 'SCIPY' or (
                self.Alg[-5:]
            ).upper == 'SCIPY':
                self.SciPyAlg = True
            elif (self.Alg[:5]).upper() == 'PYGMO' or (
                self.Alg[-5:]
            ).upper == 'PYGMO':
                self.pyGmoAlg = True
            elif (self.Alg[:5]).upper() == 'NLOPT' or (
                self.Alg[-5:]
            ).upper == 'NLOPT':
                self.nloptAlg = True
            else:
                raise Exception('Not a valid optimization algorithm')
            # File handling
            if self.RunFolder:
                self.MainDir = os.getcwd()
                self.RunDir = self.MainDir + os.sep + 'DesOpt' + self.Name
                if os.path.isdir(self.RunDir):
                    datetime.time.wait(1)
                    self.t0 = datetime.datetime.now()
                    self.t0str = self.t0.strftime('%Y%m%d%H%M%S')
                    self.Name = self.ModelName + self.Alg + self.t0str
                    self.RunDir = self.MainDir + os.sep + 'DesOpt' + self.Name
                shutil.copytree(
                    self.MainDir,
                    self.RunDir,
                    ignore=shutil.ignore_patterns(
                        'DesOpt*', '.*', '.git*', '*.pyc', 'tmp*', '__*'
                    ),
                )
                os.chdir(self.RunDir)

            if self.Monitoring and self.RunFolder:
                print(
                    '\nTo start monitoring visualization,'
                    + ' enter the following command in terminal:'
                    + '\n\n    bokeh serve --show DesOptMonitor.py \n\n'
                    + 'from the following folder:'
                    '\n\n    ' + self.RunDir + '\n\n'
                )

            def SysEq(xVal):
                # create folder and change into it
                if self.SaveEvaluations:
                    EvalDir = (
                        self.RunDir + os.sep + str(self.nEval + 1).zfill(5)
                    )
                    shutil.copytree(
                        self.MainDir,
                        EvalDir,
                        ignore=shutil.ignore_patterns(
                            'DesOpt*', '.git*', '.*', '*.pyc', 'tmp*', '__*'
                        ),
                    )
                    os.chdir(EvalDir)

                # Use only physical design variables (some algorithms
                # supplement design vector with other values, e.g. CONMIN)
                xVal = xVal[0 : self.nx]

                # Denorm and assign of design variables
                if self.xVector:
                    if self.xNorm[0]:
                        xVal = denormalize(xVal, self.xL, self.xU)
                    setattr(self.Model, self.x, xVal)
                else:
                    for i, xi in enumerate(xVal):
                        if self.xNorm[i]:
                            xVal[i] = denormalize(xi, self.xL[i], self.xU[i])
                        if self.x is not None:
                            setattr(self.Model, self.x[i], xVal[i])

                # Call
                if self.x is None:
                    eval('self.' + self.Primal + '(xVal)')
                else:
                    eval('self.Model.' + self.Primal + '(self.Model)')

                # Objective and scaling (norm)
                fVal = getattr(self.Model, self.f[0])
                if self.fNorm[0]:
                    if self.f0 is None:
                        self.f0 = fVal
                    if self.f0 == 0:
                        fVal = fVal * self.fNormMultiplier
                    else:
                        fVal = fVal / abs(self.f0) * self.fNormMultiplier
                # change the above to allow for multiobjective
                # fVal = []
                # for fi in self.f:
                #    print(getattr(self, fi))

                # Formulate constraint function (upper/lower, normalized)
                rconval = []
                if self.g:
                    gVal = [None] * self.ng
                    for i, gi in enumerate(self.g):
                        rconvalAll = np.array(getattr(self.Model, gi))
                        if np.size(rconvalAll) == 1:
                            rconvalAll = rconvalAll.reshape((1,))
                        # nr = np.size(rconval)
                        # for j in range(nr):
                        for j, rconval in enumerate(rconvalAll):
                            if self.gNorm[i + j]:
                                if self.gType[i + j] == 'upper':
                                    gVal[i + j] = (
                                        rconval / self.gLimit[i + j] - 1
                                    )
                                elif self.gType[i + j] == 'lower':
                                    gVal[i + j] = (
                                        1 - rconval / self.gLimit[i + j]
                                    )
                            else:
                                if self.gType[i + j] == 'upper':
                                    gVal[i + j] = rconval - self.gLimit[i + j]
                                elif self.gType[i + j] == 'lower':
                                    gVal[i + j] = self.gLimit[i + j] - rconval
                else:
                    gVal = []
                # Revert to normalized design variables (why necessary?)
                for i in range(len(self.x0)):
                    if self.xNorm[i]:
                        xVal[i] = normalize(xVal[i], self.xL[i], self.xU[i])

                # Move back into run folder
                if self.SaveEvaluations:
                    os.chdir('..')

                self.nEval += 1

                return (fVal, gVal, 0)

            def SensEq(xVal, fVal, gVal):
                # create folder and change into it
                if self.SaveEvaluations:
                    EvalDir = (
                        self.RunDir
                        + os.sep
                        + str(self.nSensEval + 1).zfill(5)
                        + 'Sens'
                    )
                    shutil.copytree(
                        self.MainDir,
                        EvalDir,
                        ignore=shutil.ignore_patterns(
                            'DesOpt*', '.git*', '.*', '*.pyc', 'tmp*', '__*'
                        ),
                    )
                    os.chdir(EvalDir)

                xVal = xVal[0 : self.nx]

                # Denorm and assign of design variables
                if self.xVector:
                    if self.xNorm[0]:
                        xVal = denormalize(xVal, self.xL, self.xU)
                    setattr(self.Model, self.x, xVal)
                else:
                    for i, xi in enumerate(xVal):
                        if self.xNorm[i]:
                            xVal[i] = denormalize(xi, self.xL[i], self.xU[i])
                        if self.x is not None:
                            setattr(self, self.x[i], xVal[i])

                # Call
                if self.x is None:
                    eval('self.' + self.Sensitivity + '(xVal)')
                else:
                    eval('self.Model.' + self.Sensitivity + '(self.Model)')

                # Senstivity of objective
                fNablaVal = getattr(self.Model, self.fNabla[0])
                if self.fNorm[0]:
                    if self.f0 == 0:
                        fNablaVal = fNablaVal * self.fNormMultiplier
                    else:
                        fNablaVal = (
                            fNablaVal / abs(self.f0) * self.fNormMultiplier
                        )
                for i in range(len(xVal)):
                    if self.xNorm[i]:
                        fNablaVal[i] *= self.xU[i] - self.xL[i]

                # Senstivity of constraints
                # Not working for gVector type of constraint definition
                # TODO redesign for gVector = True
                if self.g:
                    gNablaVal = [None] * self.ng
                    if self.gVector:
                        rNablaValAll = getattr(self.Model, self.gNabla[0])
                    for i, gNablai in enumerate(self.gNabla):
                        # if self.gVector:
                        #     rNablaVal = rNablaValAll[i]
                        # else:
                        #     rNablaVal = getattr(self.Model, gNablai)
                        rNablaVal = getattr(self.Model, gNablai)
                        if self.gNorm[i]:
                            if self.gType[i] == 'upper':
                                gNablaVal[i] = rNablaVal / self.gLimit[i]
                            elif self.gType[i] == 'lower':
                                gNablaVal[i] = -rNablaVal / self.gLimit[i]
                        else:
                            if self.gType[i] == 'upper':
                                gNablaVal[i] = rNablaVal
                            elif self.gType[i] == 'lower':
                                gNablaVal[i] = -rNablaVal
                    for i in range(len(self.x0)):
                        if self.xNorm[i]:
                            for j in range(len(self.gNabla)):
                                gNablaVal[j][i] *= self.xU[i] - self.xL[i]
                else:
                    gNablaVal = []

                # Revert to normalized design variables (why necessary?)
                for i in range(len(self.x0)):
                    if self.xNorm[i]:
                        xVal[i] = normalize(xVal[i], self.xL[i], self.xU[i])

                # Move back into run folder
                if self.SaveEvaluations:
                    os.chdir('..')

                self.nSensEval += 1
                return (np.array([fNablaVal]), gNablaVal, 0)

            # Maybe all too much, can i just put this as a function in opt call?
            # TODO move the next 40 lines to new file
            def ObjCon(xVal):
                f, g, fail = SysEq(xVal)
                np.array([f])
                np.array(g)
                if self.g is not None:
                    f = np.concatenate((f, g))
                return f

            def AutoSensEq(xVal, fVal, gVal):
                import autograd

                for i, xi in enumerate(xVal):
                    if self.xNorm[i]:
                        xVal[i] = denormalize(xi, self.xL[i], self.xU[i])
                ObjConNabla = autograd.jacobian(ObjCon)
                fNablaVal = ObjConNabla(xVal)
                if self.g is not None:
                    fNablaVal, rNablaVal = fNablaVal[0], fNablaVal[1:]
                if self.fNorm[0]:
                    if self.f0 == 0:
                        fNablaVal = fVal * self.fNormMultiplier
                    else:
                        fNablaVal = fVal / abs(self.f0) * self.fNormMultiplier
                for i in range(len(xVal)):
                    if self.xNorm[i]:
                        fNablaVal[i] *= self.xU[i] - self.xL[i]
                if self.g:
                    gNablaVal = [None] * len(self.gNabla)
                    for i, gNablai in enumerate(self.gNabla):
                        if self.gNorm[i]:
                            if self.gType[i] == 'upper':
                                gNablaVal[i] = rNablaVal[i] / self.gLimit[i]
                            elif self.gType[i] == 'lower':
                                gNablaVal[i] = -rNablaVal[i] / self.gLimit[i]
                        else:
                            if self.gType[i] == 'upper':
                                gNablaVal[i] = rNablaVal[i]
                            elif self.gType[i] == 'lower':
                                gNablaVal[i] = -rNablaVal[i]
                    for i in range(len(self.x0)):
                        if self.xNorm[i]:
                            for j in range(len(self.gNabla)):
                                gNablaVal[j][i] *= self.xU[i] - self.xL[i]
                else:
                    gNablaVal = []
                for i in range(len(self.x0)):
                    if self.xNorm[i]:
                        xVal[i] = normalize(xVal[i], self.xL[i], self.xU[i])
                self.nSensEval += 1

            # Normalization
            x0 = [None] * self.nx
            xL = [None] * self.nx
            xU = [None] * self.nx
            for i in range(self.nx):
                if self.xNorm[i]:
                    x0[i] = normalize(self.x0[i], self.xL[i], self.xU[i])
                    xL[i] = 0
                    xU[i] = 1
                else:
                    x0[i] = self.x0[i]
                    xL[i] = self.xL[i]
                    xU[i] = self.xU[i]

            # ----------------------------------------------
            # Seperate file an dchild class??...
            if self.pyOptAlg:
                """
                PyOpt
                """
                import pyOpt

                Alg = eval('pyOpt.' + self.Alg + '()')
                Problem = pyOpt.Optimization(self.ModelName, SysEq)
                for i in range(self.nx):
                    Problem.addVar(
                        'x' + str(i + 1),
                        'c',
                        value=x0[i],
                        lower=xL[i],
                        upper=xU[i],
                    )
                for i in range(self.nf):
                    Problem.addObj('f' + str(i + 1))
                if self.g:
                    for i in range(self.ng):
                        Problem.addCon('g' + str(i + 1), 'i')
                if self.PrintOutput:
                    print(Problem)

                # Call
                if self.Sensitivity is None:
                    if self.Alg in [
                        'NLPQLP',
                        'MMA',
                        'GCMMA',
                        'IPOPT',
                        'FSQP',
                        'PSQP',
                        'CONMIN',
                        'MMFD',
                        'SLSQP',
                        'SOLVEOPT',
                    ]:
                        fOpt, xOpt, inform = Alg(
                            Problem, sens_step=self.xDelta, store_hst=self.Name
                        )
                    else:
                        fOpt, xOpt, inform = Alg(Problem, store_hst=self.Name)

                elif self.Sensitivity == 'autograd':
                    fOpt, xOpt, inform = Alg(
                        Problem, sens_type=AutoSensEq, store_hst=self.Name
                    )
                else:
                    self.nSensEval = 0
                    fOpt, xOpt, inform = Alg(
                        Problem, sens_type=SensEq, store_hst=self.Name
                    )

                # proper size xOpt
                xOpt = np.array(xOpt).reshape(
                    len(xOpt),
                )

                # proper size fOpt
                fOpt = np.array(fOpt).reshape(
                    1,
                )

                # Todo: same for all algorithms?
                # Todo change to array?
                # Denormalization
                self.xOpt = [None] * self.nx
                self.xNormOpt = xOpt
                self.fNormOpt = fOpt
                for i in range(len(self.x0)):
                    if self.xNorm[i]:
                        self.xOpt[i] = denormalize(
                            xOpt[i], self.xL[i], self.xU[i]
                        )
                    else:
                        self.xOpt[i] = xOpt[i]
                if self.fNorm[0]:
                    if self.f0 == 0:
                        self.fOpt = fOpt / self.fNormMultiplier
                    else:
                        self.fOpt = fOpt * abs(self.f0) / self.fNormMultiplier
                else:
                    self.fOpt = fOpt

                # if self.PrintOutput:
                #     pass

                try:
                    self.inform = Alg.getInform(0)
                except:
                    self.inform = inform

                self.readHistory()

            # ----------------------------------------------
            elif self.pyGmoAlg:
                from DesOptPy.interfaces import PyGmo

                PyGmo.OptPyGmo(self, x0, xL, xU, SysEq)

            # ----------------------------------------------
            # Seperate file and child class??
            elif self.SciPyAlg:
                from DesOptPy.interfaces import SciPy

                SciPy.OptSciPy(self, x0, xL, xU, SysEq)

            # ----------------------------------------------
            # Seperate file and child class??
            elif self.nloptAlg:
                from DesOptPy.interfaces import NlOpt

                NlOpt.OptNlOpt(self, x0, xL, xU, SysEq)

            def optimizeMutiobjective(self):
                pass

            # Optimization over...now postprocessing and such

            self.t1 = datetime.datetime.now()
            self.tOpt = self.t1 - self.t0

            if self.PrintOutput:
                printResults(self)

            if self.RunFolder:
                os.chdir(self.MainDir)
                if self.RemoveRunFolder:
                    shutil.rmtree(self.RunDir)

    return Opt()



"""
horrible solution as calling syseq 2x, one obj, one constraint
"""
import numpy as np
from DesOptPy.scaling import normalize, denormalize


def OptSciPy(self, x0, xL, xU, SysEq):
    def ObjFnSciPy(xVal):
        if np.array_equal(self.xLast, xVal) != True:
            self.fVal, self.gVal, flag = SysEq(xVal)
            self.xLast = xVal.copy()
            self.fAll.append(self.fVal)
            self.gAll.append(self.gVal)
            self.xAll.append(xVal)
        return self.fVal

    def ConFnSciPy(xVal):
        if np.array_equal(self.xLast, xVal) != True:
            self.fVal, self.gVal, flag = SysEq(xVal)
            self.xLast = xVal.copy()
            self.fAll.append(self.fVal)
            self.gAll.append(self.gVal)
            self.xAll.append(xVal)
        return self.gVal

    """
    SciPy

    TODO: options need to be mapped to main options
    TODO: constraint formulation with upper or lower

    The minimize function supports the following methods:

        minimize(method=’Nelder-Mead’)
        minimize(method=’Powell’)
        minimize(method=’CG’)
        minimize(method=’BFGS’)
        minimize(method=’Newton-CG’)
        minimize(method=’L-BFGS-B’)
        minimize(method=’TNC’)
        minimize(method=’COBYLA’)
        minimize(method=’SLSQP’)
        minimize(method=’trust-constr’)
        minimize(method=’dogleg’)
        minimize(method=’trust-ncg’)
        minimize(method=’trust-krylov’)
        minimize(method=’trust-exact’)

    """
    from scipy import optimize as spopt

    self.gAll = []
    self.fAll = []
    self.xAll = []
    if 'SLSQP' in (self.Alg).upper():
        Results = spopt.minimize(
            ObjFnSciPy,
            x0,
            args=(),
            method='SLSQP',
            jac=None,
            bounds=spopt.Bounds(xL, xU),
            constraints=spopt.NonlinearConstraint(ConFnSciPy, -np.inf, 0),
            tol=None,
            callback=None,
            options={
                'maxiter': 100,
                'ftol': 1e-06,
                'iprint': 1,
                'disp': False,
                'eps': self.xDelta,
                'finite_diff_rel_step': None,
            },
        )
    elif 'trust-constr' in (self.Alg).lower():
        Results = spopt.minimize(
            ObjFnSciPy,
            x0,
            method='trust-constr',
            bounds=spopt.Bounds(xL, xU),
            constraints=spopt.NonlinearConstraint(ConFnSciPy, -np.inf, 0),
            options={
                'xtol': 1e-08,
                'gtol': 1e-08,
                'barrier_tol': 1e-08,
                'sparse_jacobian': None,
                'maxiter': 1000,
                'verbose': 0,
                'finite_diff_rel_step': None,
                'initial_constr_penalty': 1.0,
                'initial_tr_radius': 1.0,
                'initial_barrier_parameter': 0.1,
                'initial_barrier_tolerance': 0.1,
                'factorization_method': None,
                'disp': False,
            },
        )
        #'grad': None,
    elif 'differential_evolution' in (self.Alg).lower():
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html#scipy.optimize.differential_evolutionhttps://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.differential_evolution.html#scipy.optimize.differential_evolution
        """
        Results = spopt.differential_evolution(
            ObjFnSciPy,
            x0=x0,
            bounds=spopt.Bounds(xL, xU),
            args=(),
            strategy='best1bin',
            maxiter=1000,
            popsize=15,
            tol=0.01,
            mutation=(0.5, 1),
            recombination=0.7,
            seed=None,
            disp=False,
            polish=True,
            init='latinhypercube',
            atol=0,
            updating='immediate',
            workers=1,
            constraints=spopt.NonlinearConstraint(ConFnSciPy, -np.inf, 0),
        )

    # elif 'COBYLA' in (self.Alg).upper():
    #     """
    #     Not working, no bounds
    #     https://docs.scipy.org/doc/scipy/reference/optimize.minimize-cobyla.html
    #     """
    #     Results = spopt.minimize(ObjFnSciPy, x0,
    #                              args=(),
    #                              bounds=spopt.Bounds(xL, xU),
    #                              method='COBYLA',
    #                              constraints=spopt.NonlinearConstraint(ConFnSciPy, -np.inf, 0),
    #                              tol=None,
    #                              callback=None,
    #                              options={'rhobeg': 1.0,
    #                                       'maxiter': 1000,
    #                                       'disp': False,
    #                                       'catol': 0.0002})
    #     Results.nit = None
    #     Results.njev = None
    #     Results.jac = None
    # elif 'dual_annealing' in (self.Alg).lower():
    #     """
    #     No constraint
    #     https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.dual_annealing.html
    #     """
    #     Results = spopt.dual_annealing(ObjFnSciPy,
    #                                    args=(),
    #                                    bounds=tuple((np.block([xL, xU]).T).tolist()),
    #                                    #bounds=([0, 1], [0, 1]), #(xL, xU),
    #                                    maxiter=1000,
    #                                    local_search_options={},
    #                                    initial_temp=5230.0,
    #                                    restart_temp_ratio=2e-05,
    #                                    visit=2.62,
    #                                    accept=- 5.0,
    #                                    maxfun=10000000.0,
    #                                    seed=None,
    #                                    no_local_search=False,
    #                                    callback=None,
    #                                    x0=x0)
    #     Results.jac = None
    elif 'shgo' in (self.Alg).lower():
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.shgo.html
        """
        Results = spopt.shgo(
            ObjFnSciPy,
            x0,
            args=(),
            bounds=spopt.Bounds(xL, xU),
            constraints=spopt.NonlinearConstraint(ConFnSciPy, -np.inf, 0),
            n=None,
            iters=1,
            callback=None,
            minimizer_kwargs=None,
            options=None,
            sampling_method='simplicial',
        )
    else:
        raise Exception(
            'Not a valid SciPy algorithm for constrained nonlinear optimization'
        )
    xOpt = np.array(Results.x)
    fOpt = np.array([Results.fun])

    self.xNorm0 = x0
    self.x0 = self.xAll[0]
    self.f0 = self.fAll[0]
    self.g0 = self.gAll[0]
    if (self.Alg).upper() in ['slsqp', 'trust-constr']:
        self.fNablaOpt = Results.jac
    self.nIt = Results.nit
    self.xIt = None
    self.fIt = None
    self.gIt = None

    # Todo this is ugly
    self.nIt = Results.nfev
    try:
        self.nSensEval = Results.njev
    except:
        self.nSensEval = None
    self.inform = Results.success

    if (self.Alg).upper() in ['slsqp', 'trust-constr']:
        self.fNablaOpt = Results.jac
    elif 'trust-constr' in (self.Alg).lower():
        self.fNablaOpt = Results.grad
        self.gNablaOpt = Results.jac[0]

    if self.g is not None:
        self.gMax = np.max(self.gAll, 1)

    # Denormalization
    self.xOpt = [None] * self.nx
    self.xNormOpt = xOpt
    self.fNormOpt = fOpt
    for i in range(self.nx):
        if self.xNorm[i]:
            self.xOpt[i] = denormalize(xOpt[i], self.xL[i], self.xU[i])
        else:
            self.xOpt[i] = xOpt[i]
    if self.fNorm[0]:
        self.fOpt = fOpt * self.f0 / self.fNormMultiplier
    else:
        self.fOpt = fOpt
    self.gOpt = self.gVal



import os
from pathlib import Path


this_directory = Path(__file__).parent
long_description = (this_directory / 'README.md').read_text()



def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


if __name__ == '__main__':
    from distutils.core import setup

    extra_files = package_files('DesOptPy')
    setup(
        name='DesOptPy',
        version='2022.1.2',
        description='DESign OPTimization in PYthon',
        author='E. J. Wehrle',
        author_email='Erich.Wehrle@unibz.it',
        copyright='Copyright 2015-2022 E. J. Wehrle',
        package_data={'': extra_files},
        license='GNU Lesser General Public License 3.0',
        url='https://github.com/e-dub/DesOptPy',
        packages=['DesOptPy'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    )



# -*- coding: utf-8 -*-
"""
Dixon Price test function for design optimization

xOpt =
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class DixenPrice:
    x = 10

    def SysEq(self):
        n = len(self.x)
        j = np.arange(2, n + 1)
        x2 = 2 * self.x ** 2
        self.f = np.sum(j * (x2[1:] - self.x[:-1]) ** 2) + (self.x[0] - 1) ** 2


nx = 2
OptDixenPrice = OptimizationProblem(DixenPrice)
OptDixenPrice.x = 'x'
OptDixenPrice.x0 = [0] * nx
OptDixenPrice.xL = -10
OptDixenPrice.xU = +10
OptDixenPrice.f = 'f'
OptDixenPrice.Alg = 'SLSQP'
OptDixenPrice.Primal = 'SysEq'
OptDixenPrice.optimize()
OptDixenPrice.plotConvergence()



# -*- coding: utf-8 -*-
"""
Description: Simple multi-objective test function for design optimization with use of weighting factors resulting in a Pareto front
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Multobj:
    x = 2

    def SysEq(self):
        self.f1 = self.x
        self.f2 = 100.0 - self.x ** 2
        self.f = -1 * ((gamma[ii] * self.f1) + self.f2)


nPareto = 21
xOpt = [[]] * nPareto
fOpt = [[]] * nPareto
f1 = [[]] * nPareto
f2 = [[]] * nPareto
nEval = [[]] * nPareto
gamma = np.logspace(-1, 1, nPareto)
for ii in range(nPareto):
    OptMultobj = OptimizationProblem(Multobj)
    OptMultobj.Primal = 'SysEq'
    OptMultobj.x = 'x'
    OptMultobj.xL = np.ones((1,)) * 0
    OptMultobj.xU = np.ones((1,)) * 10
    OptMultobj.x0 = np.average((OptMultobj.xL, OptMultobj.xU), 0)
    OptMultobj.f = 'f'
    OptMultobj.optimize()
    xOpt[ii] = OptMultobj.xOpt
    OptMultobj.x0 = xOpt[ii]  # Better start value
    Eval = Multobj()
    Eval.x = xOpt[ii][0]
    Eval.SysEq()
    f1[ii] = Eval.f1
    f2[ii] = Eval.f2
ParetoFront = np.array([f1, f2])
print(ParetoFront)



# -*- coding: utf-8 -*-
"""
Michalewicz test function for design optimization

xOpt = 0
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Michalewicz:
    x = np.zeros(
        [
            2,
        ]
    )

    def SysEq(self):
        michalewicz_m = 2
        n = len(self.x)
        j = np.arange(1.0, n + 1)
        self.f = -np.sum(
            np.sin(self.x)
            * np.sin(j * self.x ** 2 / np.pi) ** (2.0 * michalewicz_m)
        )


OptMichalewicz = OptimizationProblem(Michalewicz)
OptMichalewicz.Primal = 'SysEq'
OptMichalewicz.x = 'x'
OptMichalewicz.x0 = (
    np.ones(
        2,
    )
    * +0
)
OptMichalewicz.xL = (
    np.ones(
        2,
    )
    * -5
)
OptMichalewicz.xU = (
    np.ones(
        2,
    )
    * +5
)
OptMichalewicz.f = 'f'
OptMichalewicz.Alg = 'PSQP'
OptMichalewicz.optimize()
OptMichalewicz.plotConvergence()



import matplotlib.pyplot as plt
import numpy as np
import dufte
import seaborn as sns


plt.rcParams.update(
    {
        'text.usetex': True,
        'font.family': 'serif',
        'font.serif': ['Palatino'],
    }
)

maxDots = 20
maxLineLables = 25
lineThick = 0.75


def plotConvergence(
    self,
    show=True,
    savePDF=False,
    savePNG=False,
    saveSVG=False,
    saveTikZ=False,
):

    # TDO: if nEval then only dots!
    # TODO change x label!
    def ConvergencePlot(r, legend, ylabel=[], plotType=1, xAxis='it'):
        if self.xIt is None:
            xAxisVal = self.nEval
            xLabel = 'evaluation $i_{\\mathrm{eval}}$'
        else:
            xAxisVal = self.nIt
            xLabel = 'iteration $i_{it}$'
        if plotType == 1:
            fig, ax = plt.subplots()
            # if legend == 'g' and (np.max(r) - np.min(r)) > 1e3:
            #     log = True
            if np.size(r) > xAxisVal:
                for i in range(np.size(r) // xAxisVal):
                    ax.plot(
                        np.array(r)[:, i],
                        linestyle='-',
                        clip_on=False,
                        linewidth=lineThick,
                        zorder=1,
                        label='$' + legend + '_{' + str(i + 1) + '}$',
                    )
                    if xAxisVal < maxDots:
                        ax.scatter(
                            list(range(0, xAxisVal, 1)),
                            np.array(r)[:, i],
                            color='white',
                            s=50,
                            zorder=2,
                        )
                        ax.scatter(
                            list(range(0, xAxisVal, 1)),
                            np.array(r)[:, i],
                            clip_on=False,
                            s=5,
                            zorder=3,
                        )
            else:
                ax.plot(
                    r,
                    # ".",
                    linestyle='-',
                    clip_on=False,
                    linewidth=lineThick,
                    zorder=1,
                    label='$' + legend + '$',
                )
                if xAxisVal < maxDots:
                    ax.scatter(
                        list(range(0, xAxisVal, 1)),
                        r,
                        color='white',
                        s=50,
                        zorder=2,
                    )
                    ax.scatter(
                        list(range(0, xAxisVal, 1)),
                        r,
                        s=5,
                        zorder=3,
                        clip_on=False,
                    )
            plt.xlabel(xLabel)
            dufte.ylabel(ylabel)
            if np.size(r) / xAxisVal < maxLineLables + 1:
                dufte.legend()
            sns.despine()

            # axes shift
            buffer = 0.1
            xDelta = xAxisVal - 1
            yDelta = np.max(r) - np.min(r)

            ax.spines['bottom'].set_bounds(0, xAxisVal - 1)
            ax.spines['left'].set_bounds(np.min(r), np.max(r))
            if xAxisVal < 21:
                x_ticks = list(range(0, xAxisVal, 1))
                ax.xaxis.set_ticks(x_ticks)
            elif xAxisVal < 41:
                x_ticks = list(range(0, xAxisVal, 2))
                ax.xaxis.set_ticks(x_ticks)
            # yStep = yDelta//3
            # y_ticks = list(np.arange(np.min(r), np.max(r), int(yStep)))
            # ax.yaxis.set_ticks(y_ticks)
            # for label in ax.get_yticklabels()[::2]:
            #    label.set_visible(False)
            # else:
            #    x_ticks = list(range(0, self.nIt))
            # ax.yaxis.set_major_locator(ticker.MultipleLocator(base=25))
            # ax.tick_params(direction='in')
            ax.set_xlim([-buffer * xDelta, xAxisVal - 1])
            # ax.set_ylim([ax.get_ylim()[0], ax.get_ylim()[1]])
            ax.set_ylim([np.min(r) - buffer * yDelta, np.max(r)])

            # ticks inward
            ax.tick_params(direction='in')

            # make space
            fig.tight_layout()

            plotName = (
                str.split(self.RunDir, '/')[-1][6:]
                + str.split(ylabel, ' ')[0].title()
                + 'Convergence'
            ).replace('\n', '')

        else:

            def adjust_spines(ax, spines, color):
                for loc, spine in ax.spines.items():
                    if loc in spines:
                        spine.set_position(
                            ('outward', 24)
                        )  # outward by 10 points
                        if loc == 'bottom':
                            spine.set_color('black')
                        else:
                            spine.set_color(color)

                    else:
                        spine.set_color('none')  # don't draw spine

            fig = plt.figure(figsize=[6.2 * 1.666, 4.8])
            x = np.linspace(0, 6)
            y = 2 * np.sin(x)
            ax = fig.add_subplot()
            ax2 = ax.twinx()

            ax.plot(
                r[0],
                clip_on=False,
                linestyle='-',
                linewidth=lineThick,
                color='tab:blue',
                zorder=1,
            )
            ax2.plot(
                r[1],
                clip_on=False,
                linestyle='-',
                linewidth=lineThick,
                color='tab:red',
                zorder=1,
            )
            if xAxisVal < maxDots:
                ax.scatter(
                    list(range(0, xAxisVal, 1)),
                    r[0],
                    color='white',
                    s=50,
                    zorder=2,
                )
                ax.scatter(
                    list(range(0, xAxisVal, 1)),
                    r[0],
                    s=5,
                    zorder=3,
                    clip_on=False,
                    color='tab:blue',
                )
                ax2.scatter(
                    list(range(0, xAxisVal, 1)),
                    r[1],
                    color='white',
                    s=50,
                    zorder=2,
                )
                ax2.scatter(
                    list(range(0, xAxisVal, 1)),
                    r[1],
                    s=5,
                    zorder=3,
                    clip_on=False,
                    color='tab:red',
                )
            adjust_spines(ax, ['bottom', 'left'], 'tab:blue')
            adjust_spines(ax2, ['right'], 'tab:red')
            ax.set_xlim([0, xAxisVal - 1])
            ax2.set_xlim([0, xAxisVal - 1])
            if xAxisVal < 21:
                x_ticks = list(range(0, xAxisVal, 1))
                ax.xaxis.set_ticks(x_ticks)
                ax2.xaxis.set_ticks(x_ticks)
            elif xAxisVal < 41:
                x_ticks = list(range(0, xAxisVal, 2))
                ax.xaxis.set_ticks(x_ticks)
                ax2.xaxis.set_ticks(x_ticks)
            ax.set_ylim([min(r[0]), max(r[0])])
            ax2.set_ylim([min(np.min(r[1]), 0), max(np.max(r[1]), 0)])
            # ax2.set_ylim([np.min(np.array(r)[:,1]), np.max(np.array(r)[:,1])])
            # ax.set_ylabel("objective value $f$", rotation='horizontal', position=(0, 1.05))
            # ax2.set_ylabel("max constraint value $g_{\max}$", rotation='horizontal', position=(0, 1.09))

            ax.set_ylabel(
                'objective function value $f$',
                rotation='horizontal',
                verticalalignment='baseline',
            )
            ax.yaxis.set_label_coords(-0.3, 1.05)
            ax2.set_ylabel(
                'max constraint value $g_{\\max}$',
                rotation='horizontal',
                verticalalignment='baseline',
            )
            ax2.yaxis.set_label_coords(1.3, 1.05)

            # ax.set_ylabel("objective function value $f$", rotation='horizontal', verticalalignment='baseline', loc="top")
            # ax2.set_ylabel("max constraint value $g_{\max}$", rotation='horizontal', verticalalignment='baseline', loc="top")

            ax.set_xlabel(xLabel)
            ax.tick_params(axis='y', colors='tab:blue')
            ax2.tick_params(axis='y', colors='tab:red')
            ax.yaxis.label.set_color('tab:blue')
            ax2.yaxis.label.set_color('tab:red')

            # ticks inward
            ax.tick_params(direction='in')
            ax2.tick_params(direction='in')

            # make space
            fig.tight_layout()

            plotName = (
                str.split(self.RunDir, '/')[-1][6:]
                + 'ObjectiveMaxConstraintConvergence'
            )

        if savePNG:
            plt.savefig(
                plotName + '.png',
                dpi=400,
                transparent=True,
                bbox_inches='tight',
                pad_inches=0,
            )
        if saveTikZ:
            import tikzplotlib

            tikzplotlib.save(
                plotName + '.pgf',
                show_info=False,
                strict=False,
                extra_axis_parameters={
                    'ylabel style={rotate=90.0}',
                    'height=\\figureheight',
                    'width=\\figurewidth',
                    'xmin=0',
                    'ymin=' + str(np.min(r)),
                    'ymax=' + str(np.max(r)),
                },
            )
            # plt.savefig(plotName + '.pgf', transparent=True)
        if saveSVG:
            plt.savefig(
                plotName + '.svg',
                transparent=True,
                bbox_inches='tight',
                pad_inches=0,
            )
        if savePDF:
            plt.savefig(
                plotName + '.pdf',
                backend='pgf',
                transparent=True,
                bbox_inches='tight',
                pad_inches=0,
            )
        if show:
            plt.show()

    if self.xIt is None:
        ConvergencePlot(self.xAll, 'x', 'design variable value', xAxis='eval')
        # ConvergencePlot(self.xNormAll,
        #                "\hat{x}",
        #                "normalized\ndesign variable value",
        #                xAxis="eval")
        ConvergencePlot(
            self.fAll, 'f', 'objective function value', xAxis='eval'
        )
        if self.g is not None:
            ConvergencePlot(
                self.gAll, 'g', 'constraint function value', xAxis=self.nEval
            )
            ConvergencePlot(
                [self.fAll, self.gMax], ['f', 'g'], plotType=2, xAxis='eval'
            )

    else:
        ConvergencePlot(self.xIt, 'x', 'design variable value')
        ConvergencePlot(
            self.xNormIt, '\\hat{x}', 'normalized\ndesign variable value'
        )
        ConvergencePlot(self.fIt, 'f', 'objective function value')
        if self.g is not None:
            ConvergencePlot(self.gIt, 'g', 'constraint function value')
            ConvergencePlot([self.fIt, self.gMaxIt], ['f', 'g'], plotType=2)

    # # objective and max constraint convergence
    # fig, ax1 = plt.subplots()
    # ax2 = ax1.twinx()
    # ax1.plot(self.fIt, label="$f$", color="green")
    # dufte.legend()
    # ax2.plot(np.max(np.array(self.gIt), 1), label="$g_{\max}$", color="red")
    # plt.xlabel("iteration $i_{it}$")
    # dufte.legend()
    # #sns.despine()
    # plt.show()


def plotBeforeAfter(
    self,
    show=True,
    savePDF=False,
    savePNG=False,
    saveSVG=False,
    saveTikZ=False,
):
    def BarPlot(r0, rOpt, legend, ylabel=[], xlabelextra='', color='blue'):
        nx = len(r0)
        labels = [
            '$' + legend + '_{' + str(i) + ' }$' for i in range(1, nx + 1)
        ]
        labels = labels[::-1]

        index = list(range(1, nx + 1))
        index1 = np.array(index) + 0.1
        index2 = np.array(index) - 0.1
        if nx == 1:
            ysize = 0.1
        else:
            ysize = 0.3 + nx * 0.5
        fig, ax = plt.subplots(figsize=(6.2, ysize))

        plt.hlines(
            y=index1,
            xmin=0,
            xmax=r0[::-1],
            color='tab:' + color,
            alpha=0.25,
            linewidth=10,
            clip_on=False,
        )
        plt.hlines(
            y=index2,
            xmin=0,
            xmax=rOpt[::-1],
            color='tab:' + color,
            linewidth=10,
            clip_on=False,
        )

        ax.set_xlabel(xlabelextra + 'value')
        ax.set_ylabel(
            ylabel,
            rotation='horizontal',
            verticalalignment='baseline',
            loc='top',
        )

        ax.tick_params(axis='both', which='major')
        plt.yticks(index, labels)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)

        ax.spines['left'].set_bounds((1, len(index)))

        if xlabelextra == '':
            val = 0
        else:
            val = 1

        ax.set_xlim(
            np.min([0, np.min(r0), np.min(rOpt)]),
            np.max([0, np.max(r0), np.max(rOpt), val]),
        )

        ax.spines['left'].set_position(('outward', 24))
        ax.spines['bottom'].set_position(('outward', 12))

        # ticks inward
        ax.tick_params(direction='in')

        # make space
        fig.tight_layout()

        plotName = (
            str.split(self.RunDir, '/')[-1][6:]
            + xlabelextra.replace(' ', '').title()
            + str.split(ylabel, ' ')[0].title()
            + 'Bar'
        ).replace('\n', '')

        if savePNG:
            plt.savefig(
                plotName + '.png',
                dpi=400,
                transparent=True,
                bbox_inches='tight',
                pad_inches=0,
            )
        if saveTikZ:
            import tikzplotlib

            tikzplotlib.save(
                plotName + '.pgf',
                show_info=False,
                strict=False,
                extra_axis_parameters={
                    'ylabel style={rotate=90.0}',
                    'height=\\figureheight',
                    'width=\\figurewidth',
                },
            )
            # plt.savefig(plotName + '.pgf', transparent=True)
        if saveSVG:
            plt.savefig(
                plotName + '.svg',
                transparent=True,
                bbox_inches='tight',
                pad_inches=0,
            )
        if savePDF:
            plt.savefig(
                plotName + '.pdf',
                backend='pgf',
                transparent=True,
                bbox_inches='tight',
                pad_inches=0,
            )
        if show:
            plt.show()

    BarPlot(self.x0, self.xOpt, 'x', 'design variable')
    BarPlot(
        np.array(self.xNorm0),
        self.xNormOpt,
        '\\hat{x}',
        'design variable',
        'normalized ',
    )
    # BarPlot(self.fIt, "f", "objective function value")
    if self.g is not None:
        BarPlot(self.g0, self.gOpt, 'g', 'constraint function', color='red')


def PrintTikZInfo():
    tikzMessage = r"""
To use PGF files (TikZ) in LyX and LaTeX:

At top of document initialize values:

    \newlength\figureheight
    \newlength\figurewidth

Where you want figure, use the following code
(set height and width to desire size):

    \setlength\figureheight{5.0cm}
    \setlength\figurewidth{7cm}
    \input{FigureName.pgf}


For Tufte style add following to LaTeX/LyX preamble:

\usepackage{pgfplots}
\pgfplotsset{compat=newest}
\pgfplotsset{every axis y label/.style={at={(-0.1, 1.05)},
                                        anchor=south east,
                                        rotate=0.0},
             axis x line*=bottom,
             axis y line*=left,
             axis x line shift=17pt,
             axis y line shift=17pt,
             separate axis lines,
             clip=false}

also changed
/usr/local/lib/python3.9/dist-packages/tikzplotlib/_text.py line 51
    scaling = 1.0 * size / data["font size"]
"""
    print(tikzMessage)



# -*- coding: utf-8 -*-
"""
Griewank test function for design optimization

xOpt = []
fOpt = 0.0
"""
from DesOptPy import OptimizationProblem
import numpy as np


class Griewank:
    x = [1, 1]

    def SysEq(self):
        fr = 4000
        n = len(self.x)
        j = np.arange(1.0, n + 1)
        s = np.sum(self.x ** 2)
        p = np.prod(np.cos(self.x / np.sqrt(j)))
        self.f = s / fr - p + 1


OptGriewank = OptimizationProblem(Griewank)
OptGriewank.x = 'x'
OptGriewank.x0 = [0] * 2
OptGriewank.xL = -5
OptGriewank.xU = 5
OptGriewank.f = 'f'
OptGriewank.Primal = 'SysEq'
OptGriewank.optimize()
OptGriewank.plotConvergence()



import numpy as np
import numpy.linalg as npla


def checkActiveConstraints(self, activeTol=1e-3):
    self.igActive = np.where(self.gOpt > -activeTol)[0].tolist()
    self.ixLActive = np.where((self.xL - self.xOpt) > -activeTol)[0].tolist()
    self.ixUActive = np.where((self.xOpt - self.xU) > -activeTol)[0].tolist()
    # if len(self.igActive) == 0:
    #     self.igActive = None
    # if len(self.ixLActive) == 0:
    #     self.ixLActive = None
    # if len(self.ixUActive) == 0:
    #     self.ixUActive = None
    # self.iActive = np.block([self.igActive, self.ixLActive, self.ixUActive])
    self.iActive = self.igActive + self.ixLActive + self.ixUActive


def LagrangianFunction(self):
    if not hasattr(self, 'iActive'):
        checkActiveConstraints(self)
    self.ConNabla = np.block(
        [self.gNablaOpt.T, np.diag([-1] * self.nx).T, np.diag([1] * self.nx).T]
    )
    self.ConActiveNabla = np.block(
        [
            self.gNablaOpt[self.igActive, :].T,
            np.diag([-1] * self.nx)[self.ixLActive, :].T,
            np.diag([1] * self.nx)[self.ixUActive, :].T,
        ]
    )
    # self.LambdaAll = npla.pinv(self.ConNabla)@-self.fNablaOpt
    # self.LambdaActiveCheck = self.LambdaAll[self.iActive]
    self.Lambda = np.zeros((self.ng + 2 * self.nx,))
    self.LambdaActive = npla.pinv(self.ConActiveNabla) @ -self.fNablaOpt
    self.Lambda[self.iActive] = self.LambdaActive
    self.OptResidual = self.fNablaOpt + self.ConNabla @ self.Lambda


def checkKKT(self, KKTTol=1e-3):
    if not hasattr(self, 'iActive'):
        checkActiveConstraints(self)
    if not hasattr(self, 'Lambda'):
        LagrangianFunction(self)
    # from numpy.linalg import norm, lstsq, pinv
    # self.kkteps = 1e-3
    # iActive = list(np.array(self.gOpt) > -self.kkteps).index(True)
    # if len(iActive) == 1:
    #    lam= np.divide(np.array(self.fNablaIt[-1]),
    #                   np.array(self.gNablaIt[-1]).reshape(5,2)[iActive,:])
    #    Lambda = np.average(lam)
    self.PrimalFeas = max(self.gOpt) < KKTTol
    self.ComplSlack = max(self.gOpt * self.Lambda[0 : self.ng]) < KKTTol
    self.DualFeas = min(self.Lambda) > -KKTTol
    self.KKTOpt = bool(self.PrimalFeas * self.DualFeas * self.ComplSlack)
    self.Opt1Order = np.linalg.norm(self.OptResidual)
    self.KKTMax = max(abs(self.OptResidual))
    if self.KKTOpt:
        print('Karush-Kuhn-Tucker optimality criteria fulfilled')
    elif self.KKTOpt == 0:
        print('Karush-Kuhn-Tucker optimality criteria NOT fulfilled')
    # if self.Opt1Order:
    print(
        'First-order residual of Lagrangian function = ' + str(self.Opt1Order)
    )


def calcShadowPrices(self):
    if not hasattr(self, 'iActive'):
        checkActiveConstraints(self)
    if not hasattr(self, 'Lambda'):
        LagrangianFunction(self)
    self.ShadowPrice = np.zeros_like(self.Lambda)
    for i in range(self.ng):
        if self.gType[i] == 'upper':
            if self.gNorm[i]:
                self.ShadowPrice[i] = -self.Lambda[i] / self.gLimit[i]
            else:
                self.ShadowPrice[i] = -self.Lambda[i]
        else:
            if self.gNorm[i]:
                self.ShadowPrice[i] = self.Lambda[i] / self.gLimit[i]
            else:
                self.ShadowPrice[i] = self.Lambda[i]
    for i in range(self.ng, self.ng + self.nx):
        self.ShadowPrice[i] = self.Lambda[i]
    for i in range(self.ng + self.nx, self.ng + self.nx + self.nx):
        self.ShadowPrice[i] = -self.Lambda[i]
    self.gShadowPrice = self.ShadowPrice[0 : self.ng]
    self.xLShadowPrice = self.ShadowPrice[self.ng : self.ng + self.nx]
    self.xUShadowPrice = self.ShadowPrice[
        self.ng + self.nx : self.ng + self.nx + self.nx
    ]


if __name__ == '__main__':
    print('testing')

    class test:
        pass

    self = test()
    self.gOpt = np.array([0, 1, -1])
    self.fOpt = np.array([10])
    self.xOpt = np.array([3, 2, 3])
    self.xL = np.array([1, 1, 1])
    self.xU = np.array([3, 3, 3])
    checkActiveConstraints(self)
